# Copyright (c) 2020 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import itertools
import signal
import time
import traceback

import testtools

import tobiko


class ITimerTest(testtools.TestCase):

    do_wait = True

    def test_itimer(self, signal_number=None, delay=0.01,
                    exception=tobiko.ITimerExpired, on_timeout=None):
        with tobiko.itimer(delay=delay, signal_number=signal_number,
                           on_timeout=on_timeout):
            if exception is None:
                self._wait_for_timeout()
            else:
                ex = self.assertRaises(exception, self._wait_for_timeout)
                self.assertEqual(signal_number or signal.SIGALRM,
                                 ex.signal_number)
                self.assertIn('_wait_for_timeout', ex.stack)

    def test_itimer_with_sigalrm(self):
        self.test_itimer(signal_number=signal.SIGALRM)

    def test_itimer_with_sigvtalrm(self):
        self.test_itimer(signal_number=signal.SIGVTALRM)

    def test_itimer_with_sigprof(self):
        self.test_itimer(signal_number=signal.SIGPROF)

    def test_itimer_with_on_timeout(self):

        counter = itertools.count()

        def on_timeout(signal_number, frame):
            self.do_wait = False
            next(counter)
            self.assertEqual(signal.SIGALRM, signal_number)
            stack = traceback.extract_stack(frame, 1)
            self.assertEqual('_wait_for_timeout', stack[0].name)

        self.test_itimer(signal_number=signal.SIGALRM,
                         on_timeout=on_timeout,
                         exception=None)
        self.assertEqual(1, next(counter))

    def _wait_for_timeout(self):
        while self.do_wait:
            time.sleep(0.)



from __future__ import absolute_import

import time

from oslo_log import log
import pandas
import six

import tobiko
from tobiko.tripleo import overcloud
from tobiko.shell import sh
from tobiko.openstack import topology


LOG = log.getLogger(__name__)


class PcsResourceException(tobiko.TobikoException):
    message = "pcs cluster is not in a healthy state"


def get_random_controller_ssh_client():
    """get a random controler's ssh client """
    nodes = topology.list_openstack_nodes(group='controller')
    controller_node = nodes[0]
    return controller_node.ssh_client


def get_pcs_resources_table(timeout=720, interval=2):
    """
    get pcs status from a controller and parse it
    to have it's resources states in check
       returns :
       rabbitmq-bundle-0    (ocf::heartbeat:rabbitmq-cluster):      Started con
       troller-0
     ip-10.0.0.101  (ocf::heartbeat:IPaddr2):       Started controller-1
       openstack-cinder-volume-docker-0     (ocf::heartbeat:docker):        Sta
       rted controller-0

    :return: dataframe of pcs resources stats table
    """
    failures = []
    start = time.time()

    ssh_client = get_random_controller_ssh_client()

    # prevent pcs table read failure while pacemaker is starting
    while time.time() - start < timeout:
        failures = []
        try:
            output = sh.execute("sudo pcs status resources |grep ocf",
                                ssh_client=ssh_client,
                                expect_exit_status=None).stdout
            # remove the first column when it only includes '*' characters
            output = output.replace('*', '').strip()
            stream = six.StringIO(output)
            table = pandas.read_csv(stream, delim_whitespace=True, header=None)
            table.columns = ['resource', 'resource_type', 'resource_state',
                             'overcloud_node']
        except ValueError:
            pcs_status_raw = sh.execute("sudo pcs status ",
                                        ssh_client=ssh_client,
                                        expect_exit_status=None).stdout
            failures.append(f'pcs status table import failed : '
                            f'pcs status stdout:\n {pcs_status_raw}')
            LOG.info('Retrying , timeout at: {}'
                     .format(timeout-(time.time() - start)))
            time.sleep(interval)
        else:
            break
    # exhausted all retries
    if failures:
        tobiko.fail(
            'pcs status table import error\n{!s}', '\n'.join(failures))

    LOG.debug("Got pcs status :\n%s", table)
    return table


class PacemakerResourcesStatus(object):
    """
    class to handle pcs resources checks
    """
    def __init__(self):
        self.pcs_df = get_pcs_resources_table()

    def container_runtime(self):
        if not self.pcs_df[(self.pcs_df['resource_type'] ==
                            "(ocf::heartbeat:docker):")].empty:
            return 'docker'
        if not self.pcs_df[(self.pcs_df['resource_type'] ==
                            "(ocf::heartbeat:podman):")].empty:
            return 'podman'

    def resource_count(self, resource_type):
        return self.pcs_df[(self.pcs_df['resource_type'] == resource_type)][
            'resource_state'].count()

    def resource_count_in_state(self, resource_type, resource_state):
        return self.pcs_df[(self.pcs_df['resource_type'] ==
                            resource_type) & (self.pcs_df['resource_state'] ==
                                              resource_state)][
            'resource_state'].count()

    def rabbitmq_resource_healthy(self):
        nodes_num = self.resource_count("(ocf::heartbeat:rabbitmq-cluster):")
        started_num = self.resource_count_in_state(
            "(ocf::heartbeat:rabbitmq-cluster):", "Started")
        if nodes_num == started_num:
            LOG.info("pcs status check: resource rabbitmq is in healthy state")
            return True
        else:
            LOG.info("pcs status check: resource rabbitmq not in healthy "
                     "state")
            return False

    def galera_resource_healthy(self):
        nodes_num = self.resource_count("(ocf::heartbeat:galera):")
        master_num = self.resource_count_in_state("(ocf::heartbeat:galera):",
                                                  "Master")
        if nodes_num == master_num:
            LOG.info("pcs status check: resource galera is in healthy state")
            return True
        else:
            LOG.info("pcs status check: resource galera not in healthy state")
            return False

    def redis_resource_healthy(self):
        nodes_num = self.resource_count("(ocf::heartbeat:redis):")
        master_num = self.resource_count_in_state(
            "(ocf::heartbeat:redis):", "Master")
        slave_num = self.resource_count_in_state(
            "(ocf::heartbeat:redis):", "Slave")
        if (master_num == 1) and (slave_num == nodes_num - master_num):
            LOG.info("pcs status check: resource redis is in healthy state")
            return True
        else:
            LOG.info("pcs status check: resource redis not in healthy state")
            return False

    def vips_resource_healthy(self):
        nodes_num = self.resource_count("(ocf::heartbeat:IPaddr2):")
        started_num = self.resource_count_in_state(
            "(ocf::heartbeat:IPaddr2):", "Started")
        if nodes_num == started_num:
            LOG.info("pcs status check: resources vips are in healthy state")
            return True
        else:
            LOG.info(
                "pcs status check: resources vips are not in healthy state")
            return False

    def ha_proxy_cinder_healthy(self):

        nodes_num = self.resource_count("(ocf::heartbeat:{}):".format(
            self.container_runtime()))
        started_num = self.resource_count_in_state(
            "(ocf::heartbeat:{}):".format(self.container_runtime()), "Started")
        if nodes_num == started_num:
            LOG.info("pcs status check: resources ha_proxy and"
                     " cinder are in healthy state")
            return True
        else:
            LOG.info(
                "pcs status check: resources ha_proxy and cinder are not in "
                "healthy state")
            return False

    def ovn_resource_healthy(self):
        if self.pcs_df.query(
                'resource_type == "(ocf::ovn:ovndb-servers):"').empty:
            LOG.info('pcs status check: ovn is not deployed, skipping ovn '
                     'resource check')
            return True
        nodes_num = self.resource_count("(ocf::ovn:ovndb-servers):")
        master_num = self.resource_count_in_state(
            "(ocf::ovn:ovndb-servers):", "Master")
        slave_num = self.resource_count_in_state(
            "(ocf::ovn:ovndb-servers):", "Slave")
        if (master_num == 1) and (slave_num == nodes_num - master_num):
            LOG.info(
                "pcs status check: resource ovn is in healthy state")
            return True
        else:
            LOG.info(
                "pcs status check: resource ovn is in not in "
                "healthy state")
            return False

    @property
    def all_healthy(self):
        """
        check if each resource is in healthy order
        and return a global healthy status
        :return: Bool
        """
        for attempt_number in range(360):

            try:

                if all([
                   self.rabbitmq_resource_healthy(),
                   self.galera_resource_healthy(),
                   self.redis_resource_healthy(),
                   self.vips_resource_healthy(),
                   self.ha_proxy_cinder_healthy(),
                   self.ovn_resource_healthy()
                   ]):
                    LOG.info("pcs status checks: all resources are"
                             " in healthy state")
                    return True
                else:

                    LOG.info("pcs status check: not all resources are "
                             "in healthy "
                             "state")
                    raise PcsResourceException()
            except PcsResourceException:
                # reread pcs status
                LOG.info('Retrying pacemaker resource checks attempt '
                         '{} of 360'.format(attempt_number))
                time.sleep(1)
                self.pcs_df = get_pcs_resources_table()
        # exhausted all retries
        tobiko.fail('pcs cluster is not in a healthy state')


def get_overcloud_nodes_running_pcs_resource(resource=None,
                                             resource_type=None,
                                             resource_state=None):
    """
    Check what nodes are running the specified resource/type/state
    resource/type/state: exact str of a resource name as seen in pcs status
    :return: list of overcloud nodes
    """
    pcs_df = get_pcs_resources_table()
    if resource:
        pcs_df_query_resource = pcs_df.query('resource=="{}"'.format(
                                        resource))
        return pcs_df_query_resource['overcloud_node'].unique().tolist()

    if resource_type and resource_state:
        pcs_df_query_resource_type_state = pcs_df.query(
            'resource_type=="{}" and resource_state=="{}"'.format(
                resource_type, resource_state))
        return pcs_df_query_resource_type_state[
            'overcloud_node'].unique().tolist()

    if resource_type and not resource_state:
        pcs_df_query_resource_type = pcs_df.query(
            'resource_type=="{}"'.format(resource_type))
        return pcs_df_query_resource_type['overcloud_node'].unique().tolist()


def get_resource_master_node(resource_type=None):
    get_overcloud_nodes_running_pcs_resource(
        resource_type=resource_type, resource_state='Master')


def get_ovn_db_master_node():
    return get_overcloud_nodes_running_pcs_resource(
        resource_type='(ocf::ovn:ovndb-servers):', resource_state='Master')


def get_overcloud_resource(resource_type=None,
                           resource_state=None):
    """
    Check what nodes are running the specified resource/type/state
    resource/type/state: exact str of a resource name as seen in pcs status
    :return: list of overcloud nodes
    """
    pcs_df = get_pcs_resources_table()

    if resource_type and resource_state:
        pcs_df_query_resource_type_state = pcs_df.query(
            'resource_type=="{}" and resource_state=="{}"'.format(
                resource_type, resource_state))
        return pcs_df_query_resource_type_state[
            'resource'].unique().tolist()

    if resource_type and not resource_state:
        pcs_df_query_resource_type = pcs_df.query(
            'resource_type=="{}"'.format(resource_type))
        return pcs_df_query_resource_type['resource'].unique().tolist()


def instanceha_deployed():
    """check IHA deployment
    checks for existence of the nova-evacuate resource"""
    if overcloud.has_overcloud():
        return get_overcloud_nodes_running_pcs_resource(
            resource='nova-evacuate')
    else:
        return False


skip_if_instanceha_not_delpoyed = tobiko.skip_unless(
    'instanceha not delpoyed', instanceha_deployed)


def fencing_deployed():
    """check fencing deployment
    checks for existence of the stonith-fence type resources"""
    ssh_client = get_random_controller_ssh_client()
    fencing_output = sh.execute("sudo pcs status |grep "
                                "'stonith:fence_ipmilan'",
                                ssh_client=ssh_client,
                                expect_exit_status=None)

    if fencing_output.exit_status == 0:
        return True
    else:
        return False


skip_if_fencing_not_deployed = tobiko.skip_unless(
    'fencing not delpoyed', fencing_deployed)



# Copyright 2020 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import collections
import time

from oslo_log import log
import yaml

import tobiko
from tobiko.shell import sh

LOG = log.getLogger(__name__)


def user_data(*args, **kwargs):
    config = cloud_config(*args, **kwargs)
    if config:
        return config.user_data
    else:
        return ''


def cloud_config(*args, **kwargs):
    return combine_cloud_configs(args + (kwargs,))


def combine_cloud_configs(objs):
    packages = []
    runcmd = []
    extra_params = {}
    for obj in objs:
        if obj:
            if not isinstance(obj, collections.abc.Mapping):
                obj = dict(obj)
            for package in obj.pop('packages', []):
                if package and package not in packages:
                    packages.append(package)
            for cmdline in obj.pop('runcmd', []):
                if cmdline:
                    cmdline = list(sh.shell_command(cmdline))
                    if cmdline:
                        runcmd.append(cmdline)
            if obj:
                extra_params.update(obj)

    return CloudConfig.create(packages=packages or None,
                              runcmd=runcmd or None,
                              **extra_params)


class CloudConfig(dict):

    @classmethod
    def create(cls, **kwargs):
        return cls((k, v)
                   for k, v in kwargs.items()
                   if v is not None)

    @property
    def user_data(self):
        return '#cloud-config\n' + yaml.dump(dict(self))

    def __add__(self, other):
        return combine_cloud_configs([self, other])


class WaitForCloudInitTimeoutError(tobiko.TobikoException):
    message = ("after {enlapsed_time} seconds cloud-init status of host "
               "{hostname!r} is still {actual!r} while it is expecting to "
               "be in {expected!r}")


def get_cloud_init_status(ssh_client=None, timeout=None):
    output = sh.execute('cloud-init status',
                        ssh_client=ssh_client,
                        timeout=timeout,
                        sudo=True).stdout
    return yaml.load(output)['status']


def wait_for_cloud_init_done(ssh_client=None, timeout=None,
                             sleep_interval=None):
    return wait_for_cloud_init_status(expected={'done'},
                                      ssh_client=ssh_client,
                                      timeout=timeout,
                                      sleep_interval=sleep_interval)


def wait_for_cloud_init_status(expected, ssh_client=None, timeout=None,
                               sleep_interval=None):
    expected = set(expected)
    timeout = timeout and float(timeout) or 1200.
    sleep_interval = sleep_interval and float(sleep_interval) or 5.
    start_time = time.time()
    actual = get_cloud_init_status(ssh_client=ssh_client, timeout=timeout)
    while actual not in expected:
        enlapsed_time = time.time() - start_time
        if enlapsed_time >= timeout:
            raise WaitForCloudInitTimeoutError(hostname=ssh_client.hostname,
                                               actual=actual,
                                               expected=expected,
                                               enlapsed_time=enlapsed_time)

        LOG.debug("Waiting cloud-init status on host %r to switch from %r to "
                  "%r...",
                  ssh_client.hostname, actual, expected)
        time.sleep(sleep_interval)
        actual = get_cloud_init_status(ssh_client=ssh_client,
                                       timeout=timeout-enlapsed_time)
    return actual



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import os
import random
import string

import testtools

import tobiko
from tobiko.openstack import heat


TEMPLATE_DIRS = [os.path.dirname(__file__)]


def random_string(length, letters=string.ascii_lowercase):
    return ''.join(random.choice(letters) for _ in range(length))


class MyStack(heat.HeatStackFixture):

    template = heat.heat_template_file(template_file='hot/my_stack.yaml',
                                       template_dirs=TEMPLATE_DIRS)

    input_text = random_string(8)


class HeatStackFixtureTest(testtools.TestCase):

    stack = tobiko.required_setup_fixture(MyStack)

    def test_get_stack(self):
        self.stack.wait_for_create_complete()
        stack = self.stack.get_stack()
        self.assertIsNotNone(stack)
        self.assertEqual(tobiko.get_fixture_name(MyStack), stack.stack_name)
        self.assertIsInstance(stack.id, str)
        self.assertIsInstance(stack.stack_status, str)

    def test_get_fixture_with_fixture_id_0(self):
        fixture_0 = tobiko.get_fixture(MyStack, fixture_id=0)
        self.assertIs(fixture_0, self.stack)

    def test_get_fixture_with_fixture_id_1(self):
        fixture_0 = tobiko.get_fixture(MyStack)
        fixture_1 = tobiko.get_fixture(MyStack, fixture_id=1)
        self.assertIsNot(fixture_0, fixture_1)
        stack_0 = tobiko.setup_fixture(fixture_0).get_stack()
        stack_1 = tobiko.setup_fixture(fixture_1).get_stack()
        self.assertNotEqual(stack_0.id, stack_1.id)
        self.assertEqual(tobiko.get_fixture_name(MyStack), stack_0.stack_name)
        self.assertEqual(tobiko.get_fixture_name(MyStack) + '-1',
                         stack_1.stack_name)



# Copyright 2020 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import io
import os

import yaml

import tobiko
from tobiko.shell import sh
from tobiko.tripleo import _undercloud
from tobiko.tripleo import _config


def get_tripleo_ansible_inventory():
    inventory_file = get_tripleo_ansible_inventory_file()
    with io.open(inventory_file, 'rb') as fd:
        return yaml.safe_load(fd)


def has_tripleo_ansible_inventory():
    inventory_file = get_tripleo_ansible_inventory_file()
    return inventory_file and os.path.isfile(inventory_file)


skip_if_missing_tripleo_ansible_inventory = \
    tobiko.skip_unless("Can't read TripleO Ansible inventory",
                       has_tripleo_ansible_inventory)


def get_tripleo_ansible_inventory_file():
    return tobiko.setup_fixture(TripleoAnsibleInventoryFixture).inventory_file


class TripleoAnsibleInventoryFixture(tobiko.SharedFixture):

    inventory_file = None

    def setup_fixture(self):
        tripleo = _config.get_tripleo_config()
        self.inventory_file = inventory_file = tobiko.tobiko_config_path(
            tripleo.inventory_file)
        if inventory_file and not os.path.isfile(inventory_file):
            content = read_tripleo_ansible_inventory()
            with io.open(inventory_file, 'w') as fd:
                fd.write(content)


READ_TRIPLEO_ANSIBLE_INVENTORY_SCRIPT = """
source {undercloud_rcfile} || exit 1

set -x

INVENTORY_FILE=$(mktemp tripleo-hosts-XXXXXXXXXX.yaml)
tripleo-ansible-inventory --ansible_ssh_user "{overcloud_ssh_username}" \\
                          --static-yaml-inventory "$INVENTORY_FILE"
RC=$?

if [ $RC == 0 ]; then
    cat "$INVENTORY_FILE"
fi
rm -fR "$INVENTORY_FILE"
exit $RC
"""


def read_tripleo_ansible_inventory():
    tripleo = _config.get_tripleo_config()
    ssh_client = _undercloud.undercloud_ssh_client()
    script = READ_TRIPLEO_ANSIBLE_INVENTORY_SCRIPT.format(
        undercloud_rcfile=tripleo.undercloud_rcfile,
        overcloud_ssh_username=tripleo.overcloud_ssh_username)
    return sh.execute('/bin/bash', stdin=script, ssh_client=ssh_client).stdout



# Copyright (c) 2019 Red Hat
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import testtools

import tobiko
from tobiko import config
from tobiko.shell import ping
from tobiko.shell import ip
from tobiko.openstack import neutron
from tobiko.openstack import stacks
from tobiko.openstack import topology


CONF = config.CONF


class RouterTest(testtools.TestCase):
    """Test Neutron routers"""

    #: Resources stack with Nova server to send messages to
    stack = tobiko.required_setup_fixture(stacks.CirrosServerStackFixture)

    def setUp(self):
        super(RouterTest, self).setUp()
        if not self.stack.network_stack.has_gateway:
            tobiko.skip_test(
                f"Stack {self.stack.network_stack.stack_name} has no gateway")

    @property
    def router(self):
        return self.stack.network_stack.gateway_details

    @property
    def ipv4_subnet_gateway_ip(self):
        if not self.stack.network_stack.has_ipv4:
            tobiko.skip_test(f"Stack {self.stack.network_stack.stack_name} "
                             "has no ipv4 subnet")
        return self.stack.network_stack.ipv4_subnet_gateway_ip

    @property
    def ipv6_subnet_gateway_ip(self):
        if not self.stack.network_stack.has_ipv6:
            tobiko.skip_test(f"Stack {self.stack.network_stack.stack_name} "
                             "has no ipv6 subnet")
        return self.stack.network_stack.ipv6_subnet_gateway_ip

    @property
    def external_gateway_ips(self):
        return self.stack.network_stack.external_gateway_ips

    @property
    def router_ips(self):
        return tobiko.select([self.ipv4_subnet_gateway_ip,
                              self.ipv6_subnet_gateway_ip] +
                             self.external_gateway_ips)

    @tobiko.retry_test_case(interval=30.)
    def test_internal_router_ipv4_interface_is_reachable(self):
        ping.assert_reachable_hosts([self.ipv4_subnet_gateway_ip],
                                    ssh_client=self.stack.ssh_client)

    @tobiko.retry_test_case(interval=30.)
    def test_internal_router_ipv6_interface_is_reachable(self):
        ping.assert_reachable_hosts([self.ipv6_subnet_gateway_ip],
                                    ssh_client=self.stack.ssh_client)

    def test_ipv4_subnet_gateway_ip(self):
        self.assertEqual(4, self.ipv4_subnet_gateway_ip.version)
        self.assertIn(self.ipv4_subnet_gateway_ip,
                      self.stack.network_stack.ipv4_gateway_addresses)

    def test_ipv6_subnet_gateway_ip(self):
        self.assertEqual(6, self.ipv6_subnet_gateway_ip.version)
        self.assertIn(self.ipv6_subnet_gateway_ip,
                      self.stack.network_stack.ipv6_gateway_addresses)

    @neutron.skip_if_missing_networking_extensions('l3_agent_scheduler')
    def test_router_is_scheduled_on_l3_agents(self):
        router_agent = neutron.find_l3_agent_hosting_router(self.router['id'],
                                                            unique=True)
        self._check_routers_namespace_on_host(router_agent['host'])

    def _check_routers_namespace_on_host(self, hostname, state="master"):
        router_namespace = "qrouter-%s" % self.router['id']
        agent_host = topology.get_openstack_node(hostname=hostname)
        namespaces = ip.list_network_namespaces(
            ssh_client=agent_host.ssh_client)
        self.assertIn(router_namespace, namespaces)
        namespace_ips = ip.list_ip_addresses(
            scope='global', network_namespace=router_namespace,
            ssh_client=agent_host.ssh_client)
        missing_ips = set(self.router_ips) - set(namespace_ips)
        if state == "master":
            self.assertFalse(missing_ips)
        else:
            self.assertTrue(missing_ips)


@neutron.skip_if_missing_networking_extensions('l3-ha')
@neutron.skip_if_missing_networking_agents(binary='neutron-l3-agent',
                                           count=2)
class L3HARouterTest(RouterTest):
    """Test Neutron HA routers"""

    #: Resources stack with Nova server to send messages to
    stack = tobiko.required_setup_fixture(stacks.L3haServerStackFixture)

    @neutron.skip_if_missing_networking_extensions('l3_agent_scheduler')
    def test_router_is_scheduled_on_l3_agents(self):
        router_agents = neutron.list_l3_agent_hosting_routers(
            self.router['id'])
        master_agent = router_agents.with_items(ha_state='active').unique
        backup_agents = router_agents.with_items(ha_state='standby')
        self.assertGreaterEqual(len(backup_agents), 1)
        self._check_routers_namespace_on_host(master_agent['host'])
        for backup_agent in backup_agents:
            self._check_routers_namespace_on_host(
                backup_agent['host'], state="backup")



# Copyright 2018 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import functools
import inspect
import typing  # noqa

import fixtures
import testtools


SkipException = testtools.TestCase.skipException  # type: typing.Type

SkipTarget = typing.Union[typing.Callable,
                          typing.Type[testtools.TestCase],
                          typing.Type[fixtures.Fixture]]
SkipDecorator = typing.Callable[[SkipTarget], SkipTarget]


def skip_test(reason: str):
    """Interrupt test case execution marking it as skipped for given reason"""
    raise SkipException(reason)


def skip(reason: str) -> SkipDecorator:
    """Mark test case for being skipped for a given reason"""
    return _skip_unless(reason, None, True)


def skip_if(reason: str, function: typing.Callable, *args, **kwargs) -> \
        SkipDecorator:
    """Mark test case for being skipped for a given reason if it matches"""
    return _skip_unless(reason, function, False, *args, **kwargs)


def skip_unless(reason: str, function: typing.Callable, *args, **kwargs) -> \
        SkipDecorator:
    """Mark test case for being skipped for a given reason unless it matches"""
    return _skip_unless(reason, function, True, *args, **kwargs)


def _skip_unless(reason: str, function: typing.Optional[typing.Callable],
                 unless: bool, *args, **kwargs) -> SkipDecorator:
    """Mark test case for being skipped for a given reason unless it matches"""

    def decorator(obj: SkipTarget) -> SkipTarget:
        method = _get_decorated_method(obj)

        @functools.wraps(method)
        def skipping_unless(*_args, **_kwargs):
            if function is not None:
                return_value = function(*args, **kwargs)
                if unless is bool(return_value):
                    return method(*_args, **_kwargs)
                if '{return_value' in reason:
                    skip_test(reason.format(return_value=return_value))
            skip_test(reason)

        if obj is method:
            return skipping_unless
        else:
            setattr(obj, method.__name__, skipping_unless)
            return obj

    return decorator


def _get_decorated_method(obj: typing.Any) -> typing.Callable:
    if inspect.isclass(obj):
        setup_method = getattr(obj, 'setUp', None)
        if callable(setup_method):
            return setup_method
        else:
            raise TypeError(f"Class {obj} does not implement setUp method")
    elif callable(obj):
        return obj
    else:
        raise TypeError(f"Object {obj} is not a class or a function")



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import socket

import six
import testtools

import tobiko
from tobiko.shell import sh
from tobiko.shell import ssh
from tobiko.openstack import keystone
from tobiko.openstack import stacks


@keystone.skip_unless_has_keystone_credentials()
class HostnameTest(testtools.TestCase):

    server_stack = tobiko.required_setup_fixture(
        stacks.CirrosServerStackFixture)

    def test_hostname(self, expect_hostname=None, **execute_params):
        hostname = sh.get_hostname(**execute_params)
        self.assertIsInstance(hostname, six.string_types)
        if expect_hostname:
            self.assertEqual(expect_hostname, hostname)
        else:
            self.assertNotEqual('', hostname)

    def test_local_hostname(self):
        self.test_hostname(expect_hostname=socket.gethostname(),
                           ssh_client=False)

    def test_remote_hostname(self, ssh_client=None):
        ssh_client = ssh_client or self.server_stack.ssh_client
        stdin, stdput, stderr = ssh_client.connect().exec_command('hostname')
        stdin.close()
        self.assertEqual(b'', stderr.read())
        self.test_hostname(ssh_client=ssh_client,
                           expect_hostname=stdput.read().decode().strip())

    def test_proxy_hostname(self):
        ssh_client = ssh.ssh_proxy_client()
        if ssh_client is None:
            self.skip('SSH proxy server not configured')
        self.test_remote_hostname(ssh_client=ssh_client)



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from heatclient.v1 import client as heatclient
import mock

from tobiko.openstack import keystone
from tobiko.openstack import heat
from tobiko.tests.unit import openstack
from tobiko.tests.unit.openstack import test_client


MockClient = mock.create_autospec(heatclient.Client)


class HeatClientFixtureTest(test_client.OpenstackClientFixtureTest):

    def create_client(self, session=None):
        return heat.HeatClientFixture(session=session)


class GetHeatClientTest(openstack.OpenstackTest):

    def test_get_heat_client(self, session=None, shared=True):
        client1 = heat.get_heat_client(session=session, shared=shared)
        client2 = heat.get_heat_client(session=session, shared=shared)
        if shared:
            self.assertIs(client1, client2)
        else:
            self.assertIsNot(client1, client2)
        self.assertIsInstance(client1, heatclient.Client)
        self.assertIsInstance(client2, heatclient.Client)

    def test_get_heat_client_with_not_shared(self):
        self.test_get_heat_client(shared=False)

    def test_get_heat_client_with_session(self):
        session = keystone.get_keystone_session()
        self.test_get_heat_client(session=session)


class HeatClientTest(openstack.OpenstackTest):

    def test_heat_client(self, obj=None):
        client = heat.heat_client(obj)
        self.assertIsInstance(client, heatclient.Client)
        if obj is None:
            self.assertIs(heat.default_heat_client(), client)
        elif isinstance(obj, heatclient.Client):
            self.assertIs(obj, client)
        elif isinstance(obj, heat.HeatClientFixture):
            self.assertIs(obj.client, client)

    def test_heat_client_with_client(self):
        self.test_heat_client(obj=MockClient())

    def test_heat_client_with_fixture(self):
        self.test_heat_client(obj=heat.HeatClientFixture())



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import docker

import tobiko
from tobiko.docker import _exception
from tobiko.docker import _shell
from tobiko.shell import ssh


def get_docker_client(base_urls=None, ssh_client=None):
    return DockerClientFixture(base_urls=base_urls,
                               ssh_client=ssh_client)


def list_docker_containers(client=None, **kwargs):
    try:
        containers = docker_client(client).containers.list(all=True, **kwargs)
    except _exception.DockerUrlNotFoundError:
        return tobiko.Selection()
    else:
        return tobiko.select(containers)


def docker_client(obj=None):
    if obj is None:
        obj = get_docker_client()
    if tobiko.is_fixture(obj):
        obj = tobiko.setup_fixture(obj).client
    if isinstance(obj, docker.DockerClient):
        return obj
    raise TypeError('Cannot obtain a DockerClient from {!r}'.format(obj))


class DockerClientFixture(tobiko.SharedFixture):

    base_urls = None
    client = None
    ssh_client = None

    def __init__(self, base_urls=None, ssh_client=None):
        super(DockerClientFixture, self).__init__()
        if base_urls:
            self.base_urls = list(base_urls)
        if ssh_client:
            self.ssh_client = ssh_client

    def setup_fixture(self):
        self.setup_ssh_client()
        self.setup_base_urls()
        self.setup_client()

    def setup_ssh_client(self):
        ssh_client = self.ssh_client
        if ssh_client is None:
            self.ssh_client = ssh_client = ssh.ssh_proxy_client() or False
            if ssh_client:
                tobiko.setup_fixture(ssh_client)
        return ssh_client

    def setup_base_urls(self):
        base_urls = self.base_urls
        if base_urls is None:
            self.base_urls = base_urls = self.discover_docker_urls()
        return base_urls

    def setup_client(self):
        client = self.client
        if client is None:
            self.client = client = self.create_client()
        return client

    def create_client(self):
        exc_info = None
        for base_url in self.base_urls:
            if self.ssh_client:
                base_url = ssh.get_port_forward_url(ssh_client=self.ssh_client,
                                                    url=base_url)
            client = docker.DockerClient(base_url=base_url)
            try:
                client.ping()
            except Exception:
                exc_info = exc_info or tobiko.exc_info()
            else:
                return client

        if exc_info:
            exc_info.reraise()
        else:
            raise _exception.DockerError('Unable to create docker client')

    def connect(self):
        return tobiko.setup_fixture(self).client

    def discover_docker_urls(self):
        return _shell.discover_docker_urls(ssh_client=self.ssh_client)



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from tobiko.shell.sh import _command
from tobiko.shell.sh import _exception
from tobiko.shell.sh import _execute
from tobiko.shell.sh import _hostname
from tobiko.shell.sh import _io
from tobiko.shell.sh import _local
from tobiko.shell.sh import _nameservers
from tobiko.shell.sh import _process
from tobiko.shell.sh import _ps
from tobiko.shell.sh import _reboot
from tobiko.shell.sh import _ssh
from tobiko.shell.sh import _uptime


ShellCommand = _command.ShellCommand
ShellCommandType = _command.ShellCommandType
shell_command = _command.shell_command

ShellError = _exception.ShellError
ShellCommandFailed = _exception.ShellCommandFailed
ShellTimeoutExpired = _exception.ShellTimeoutExpired
ShellProcessTerminated = _exception.ShellProcessTerminated
ShellProcessNotTerminated = _exception.ShellProcessNotTerminated
ShellStdinClosed = _exception.ShellStdinClosed

execute = _execute.execute
execute_process = _execute.execute_process
ShellExecuteResult = _execute.ShellExecuteResult

HostNameError = _hostname.HostnameError
get_hostname = _hostname.get_hostname

join_chunks = _io.join_chunks

local_execute = _local.local_execute
local_process = _local.local_process
LocalShellProcessFixture = _local.LocalShellProcessFixture
LocalExecutePathFixture = _local.LocalExecutePathFixture

ListNameserversFixture = _nameservers.ListNameserversFixture
list_nameservers = _nameservers.list_nameservers

process = _process.process
str_from_stream = _process.str_from_stream
ShellProcessFixture = _process.ShellProcessFixture

PsError = _ps.PsError
PsWaitTimeout = _ps.PsWaitTimeout
list_all_processes = _ps.list_all_processes
list_kernel_processes = _ps.list_kernel_processes
list_processes = _ps.list_processes
wait_for_processes = _ps.wait_for_processes

reboot_host = _reboot.reboot_host
RebootHostError = _reboot.RebootHostError
RebootHostOperation = _reboot.RebootHostOperation
RebootHostTimeoutError = _reboot.RebootHostTimeoutError
hard_reset_method = _reboot.hard_reset_method
soft_reset_method = _reboot.soft_reset_method

ssh_process = _ssh.ssh_process
ssh_execute = _ssh.ssh_execute
SSHShellProcessFixture = _ssh.SSHShellProcessFixture

get_uptime = _uptime.get_uptime



# Copyright (c) 2020 Red Hat
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import re
import typing  # noqa

from oslo_log import log
import testtools

import tobiko
from tobiko.openstack import neutron
from tobiko.openstack import nova
from tobiko.openstack import stacks
from tobiko.openstack import tests
from tobiko.openstack import topology
from tobiko.shell import ping
from tobiko.shell import sh
from tobiko.tripleo import containers
from tobiko.tripleo import overcloud


LOG = log.getLogger(__name__)

# typing hits
AgentType = typing.Dict[str, typing.Any]
AgentListType = typing.List[AgentType]


class BaseAgentTest(testtools.TestCase):

    agent_name: str = '<undefined agent name>'

    @classmethod
    def setUpClass(cls):
        cls.service_name: str = topology.get_agent_service_name(cls.agent_name)
        cls.container_name: str = ''
        cls.agents: AgentListType = \
            neutron.list_networking_agents(binary=cls.agent_name)

    def setUp(self):
        super(BaseAgentTest, self).setUp()
        if not self.agents:
            self.skipTest(f"Missing Neutron agent(s): '{self.agent_name}'")
        self.addCleanup(tests.test_neutron_agents_are_alive)

    @property
    def hosts(self) -> typing.List[str]:
        return [agent['host'] for agent in self.agents]

    @property
    def container_runtime_name(self):
        if overcloud.has_overcloud():
            return containers.container_runtime_name
        else:
            return 'docker'

    def get_ovn_agents_from_containers(self):
        if not self.agents:
            try:
                self.container_name = \
                    topology.get_agent_container_name(self.agent_name)
            except KeyError:
                LOG.debug('OVN network agents are not containerized on this'
                          'environment')
                return
            oc_containers_df = containers.list_containers_df().query(
                f'container_name == "{self.container_name}"')
            LOG.debug(
                f"{self.container_name} container found:\n{oc_containers_df}")

            self.agents = []
            for _, oc_container in oc_containers_df.iterrows():
                if oc_container['container_state'] == 'running':
                    agent_info = {'host': oc_container['container_host'],
                                  'name': oc_container['container_name']}
                    self.agents.append(agent_info)

    def stop_agent(self, hosts: typing.Optional[typing.List[str]] = None):
        '''Stop network agent on hosts

        It ensures service service is stopped and register systemd service
        restart as test case cleanup. In case of systemd service is not
        available it stops the container itself

        :parm hosts: List of hostnames to stop agent on
        :type hosts: list of strings
        '''
        hosts = hosts or self.hosts
        self.assertNotEqual([], hosts, "Host list is empty")

        for host in hosts:
            ssh_client = topology.get_openstack_node(hostname=host).ssh_client
            is_systemd = topology.check_systemd_monitors_agent(host,
                                                               self.agent_name)
            if is_systemd:
                LOG.debug(f"Stopping service '{self.service_name}' on "
                          f"host '{host}'...")
                sh.execute(f"systemctl stop {self.service_name}",
                           ssh_client=ssh_client,
                           sudo=True)
                LOG.debug(f"Service '{self.service_name}' stopped on host "
                          f"'{host}'.")
            else:
                if self.container_name == '':
                    self.container_name = topology.get_agent_container_name(
                        self.agent_name)
                LOG.debug(f'Stopping container {self.container_name} on '
                          f"host '{host}'...")
                sh.execute(f'{self.container_runtime_name} stop '
                           f'{self.container_name}',
                           ssh_client=ssh_client,
                           sudo=True)
                LOG.debug(f'Container {self.container_name} has been stopped '
                          f"on host '{host}'...")
            # Schedule auto-restart of service at the end of this test case
            self.addCleanup(self.start_agent, hosts=[host, ])

    def start_agent(self, hosts: typing.Optional[typing.List[str]] = None):
        '''Start network agent on hosts

        It ensures system service is running. If the systemd service is not
        available it starts container itself

        :parm hosts: List of hostnames to start agent on
        :type hosts: list of strings
        '''
        hosts = hosts or self.hosts
        self.assertNotEqual([], hosts, "Host list is empty")

        for host in hosts:
            ssh_client = topology.get_openstack_node(hostname=host).ssh_client
            is_systemd = topology.check_systemd_monitors_agent(host,
                                                               self.agent_name)
            if is_systemd:
                LOG.debug(f"Starting service '{self.service_name}' on "
                          f"host '{host}'...")
                sh.execute(f"systemctl start {self.service_name}",
                           ssh_client=ssh_client,
                           sudo=True)
            else:
                if self.container_name == '':
                    self.container_name = topology.get_agent_container_name(
                        self.agent_name)
                LOG.debug(f'Starting container {self.container_name} on '
                          f"host '{host}'...")
                sh.execute(f'{self.container_runtime_name} start '
                           f'{self.container_name}',
                           ssh_client=ssh_client,
                           sudo=True)

    def restart_agent_container(
            self, hosts: typing.Optional[typing.List[str]] = None):
        '''Restart network agent containers on hosts

        Restart docker or podman containers and check network agents are up and
        running after it

        :parm hosts: List of hostnames to start agent on
        :type hosts: list of strings
        '''
        hosts = hosts or self.hosts
        self.assertNotEqual([], hosts, "Host list is empty")

        self.container_name = (self.container_name or
                               topology.get_agent_container_name(
                                   self.agent_name))

        for host in hosts:
            ssh_client = topology.get_openstack_node(hostname=host).ssh_client
            sh.execute(f'{self.container_runtime_name} restart '
                       f'{self.container_name}',
                       ssh_client=ssh_client,
                       sudo=True)

    def get_cmd_pids(self, process_name, command_filter, hosts=None,
                     timeout=120, interval=2, min_pids_per_host=1) -> \
            typing.Dict[str, frozenset]:
        '''Search for PIDs that match creteria on requested hosts

        :param process_name: Name of the executable of the process
        :type process_name: string
        :parm command_filter: Regex to be found in process command details
        :type command_filter: string
        :param hosts: List of hostnames to search for processes on
        :type hosts: list of strings
        :param timeout: Time to search for processes
        :type timeout: int
        :param interval: Time to wait between searching attempts
        :param min_pids_per_host: Minimum amount of processes to be found
        :type min_pids_per_host: int
        :return: Dictionary with hostnames as a key and list of PIDs as value
        :rtype: dict
        '''
        hosts = hosts or self.hosts
        self.assertNotEqual([], hosts, "Host list is empty")

        pids_per_host = {}
        for host in hosts:
            LOG.debug(f'Search for {process_name} process on {host}')
            retry = tobiko.retry(timeout=timeout, interval=interval)
            for _ in retry:
                pids = self.list_pids(host, command_filter, process_name)
                if len(pids) >= min_pids_per_host:
                    pids_per_host[host] = frozenset(pids)
                    LOG.debug(f"Process '{process_name}' is running "
                              f"on host '{host}' (PIDs={pids!r})")
                    break
        return pids_per_host

    def list_pids(self, host, command_filter, process_name):
        '''Search for PIDs matched with filter and process name

        :param host: Hostname of the node to search processes on
        :type host: string
        :param command_filter: Regex to be found in process command details
        :type command_filter: string
        :param process_name: Name of the executable in process list
        :type process_name: string
        '''
        ssh_client = topology.get_openstack_node(hostname=host).ssh_client
        processes = sh.list_processes(command=process_name,
                                      ssh_client=ssh_client)
        pids = []
        for process in processes:
            try:
                command = sh.execute(f'cat /proc/{process.pid}/cmdline',
                                     ssh_client=ssh_client)
                if re.search(command_filter, command.stdout):
                    pids.append(process.pid)
            except sh.ShellCommandFailed:
                LOG.debug(f'Process {process.pid} has been terminated right '
                          f'after the process list has been collected')
        return pids

    def kill_pids(self, host, pids):
        '''Kill processes with specific PIDs on the host

        :param host: Hostname of the node to kill processes on
        :type host: string
        :param pids: List of PIDs to be killed
        :type pids: list of int
        '''
        ssh_client = topology.get_openstack_node(hostname=host).ssh_client
        pid_args = ' '.join(str(pid) for pid in pids)
        sh.execute(f'kill -15 {pid_args}', ssh_client=ssh_client, sudo=True)
        retry = tobiko.retry(timeout=60, interval=2)
        for _ in retry:
            pid_status = sh.execute(f'kill -0 {pid_args}',
                                    ssh_client=ssh_client,
                                    expect_exit_status=None,
                                    sudo=True).stderr.strip().split('\n')
            if all('No such process' in status for status in pid_status) and \
                    len(pid_status) == len(pids):
                break

    def wait_processes_destroyed(self, command_filter, pids_per_host,
                                 timeout=120, interval=2):
        '''Wait for processes to be terminated on hosts

        Make sure that all processes from the list are terminated or return
        an error otherwise. Tricky situation may happen when the different
        process with same PID can be spawned so then need to check it against
        `command_filter`.

        :param command_filter: Patter to be found in process command details
        :type command_filter: string
        :param pids_per_host: Dictionary with hostnames as a key and list of
                PIDs as a value
        :type pids_per_host: dict
        :param timeout: Time to wait till each process will be terminated
        :type timeout: int
        :param interval: Time to sleep between attempts
        :type interval: int
        '''
        LOG.debug(f'Waiting for processes to be finished: {pids_per_host}')
        for host, pids in pids_per_host.items():
            for pid in pids:
                retry = tobiko.retry(timeout=timeout, interval=interval)
                for _ in retry:
                    LOG.debug(f'Check if {pid} has been terminated on {host}')
                    if self.is_destroyed(pid, command_filter, host):
                        LOG.debug(f'Process {pid} finished on {host}')
                        break

    def is_destroyed(self, pid, command_filter, hostname):
        '''Check if process has been terminated

        :param pid: Process ID to check if exist and handles specific command
        :type pid: int
        :param command_filter: Patter to be found in process command details
        :type command_filter: string
        :param hostname: Hostname of the node to look for PID on
        :type hostname: string
        '''
        host = topology.get_openstack_node(hostname=hostname)
        processes = sh.list_processes(ssh_client=host.ssh_client)
        process = processes.with_attributes(pid=pid)
        process_destroyed = False
        if not process:
            LOG.debug(f'No PID {pid} has been found in process list')
            process_destroyed = True
        else:
            try:
                command = sh.execute(f'cat /proc/{pid}/cmdline',
                                     ssh_client=host.ssh_client)
                if command_filter not in command.stdout:
                    LOG.debug(f'Different process with same PID {pid} exist')
                    process_destroyed = True
            except sh.ShellCommandFailed:
                LOG.debug(f'Process {pid} has been terminated right after the'
                          f' process list has been collected')
                process_destroyed = True
        return process_destroyed


class DHCPAgentTest(BaseAgentTest):

    agent_name = neutron.DHCP_AGENT

    #: Resources stack with Nova server to send messages to
    stack = tobiko.required_setup_fixture(stacks.CirrosServerStackFixture)

    def test_stop_dhcp_agent(self):
        '''Test that dnsmasq processes are not broken after DHCP agent restart

        Dnsmasq processes should stay alive if DHCP agent is turned off and
        then restarted once DHCP agent is returned to active state.
        '''
        self.agents = neutron.list_dhcp_agent_hosting_network(
            self.stack.network)
        self.assertNotEqual(
            [], self.agents, "No DHCP agent found serving network "
            f"'{self.stack.network}'")
        pids = self.get_cmd_pids("dnsmasq", self.stack.network)

        self.stop_agent()
        self.assertEqual(pids, self.get_cmd_pids("dnsmasq",
                                                 self.stack.network))

        self.start_agent()
        self.wait_processes_destroyed(self.stack.network, pids)
        new_pids = self.get_cmd_pids("dnsmasq", self.stack.network)
        self.assertNotEqual(pids, new_pids)

    def test_dhcp_lease_served_when_dhcp_agent_down(self):
        '''Test that DHCP lease is correctly served when DHCP agent is down

        Make sure that the VM will receive IP address after the reboot.
        DHCP agent should be down during the VM reboot. VM should receive
        the same IP address that was assigned to it before the reboot.
        '''
        ping.ping_until_received(self.stack.ip_address).assert_replied()

        self.agents = neutron.list_dhcp_agent_hosting_network(
            self.stack.network)
        self.assertNotEqual(
            [], self.agents, "No DHCP agent found serving network "
            f"'{self.stack.network}'")
        pids = self.get_cmd_pids("dnsmasq", self.stack.network)
        self.stop_agent()

        nova.shutoff_server(self.stack.server_id)
        nova.activate_server(self.stack.server_id)
        ping.ping_until_received(self.stack.ip_address).assert_replied()

        self.start_agent()
        self.wait_processes_destroyed(self.stack.network, pids)
        new_pids = self.get_cmd_pids("dnsmasq", self.stack.network)
        self.assertNotEqual(pids, new_pids)


class L3AgentTest(BaseAgentTest):

    agent_name = neutron.L3_AGENT

    #: Resources stack with Nova server to send messages to
    stack = tobiko.required_setup_fixture(stacks.CirrosPeerServerStackFixture)
    ha_stack = tobiko.required_setup_fixture(stacks.L3haServerStackFixture)

    def setUp(self):
        super(L3AgentTest, self).setUp()
        self.router_id = self.stack.network_stack.gateway_id

    def wait_for_active_ha_l3_agent(self) -> AgentType:
        ha_router_id = self.ha_stack.network_stack.gateway_id
        for attempt in tobiko.retry(timeout=180., interval=5.):
            agents = neutron.list_l3_agent_hosting_routers(ha_router_id)
            try:
                active_agent = agents.with_items(ha_state='active').unique
                break
            except (tobiko.MultipleObjectsFound, tobiko.ObjectNotFound):
                attempt.check_limits()
                continue

        return active_agent

    @neutron.skip_if_missing_networking_extensions('l3-ha')
    @neutron.skip_if_missing_networking_extensions('l3_agent_scheduler')
    def test_keepalived_after_l3_agent_restart(self):
        """Verifies that keepalived survives restart of L3 agents

        Keepalived should keep the same process IDs after L3 agents have been
        restarted
        """
        self.agents = [self.wait_for_active_ha_l3_agent()]
        ha_router_id = self.ha_stack.network_stack.gateway_id
        pids = self.get_cmd_pids('keepalived', ha_router_id,
                                 min_pids_per_host=2)
        self.stop_agent()
        self.start_agent()
        self.agents = [self.wait_for_active_ha_l3_agent()]
        self.assertEqual(pids,
                         self.get_cmd_pids('keepalived',
                                           ha_router_id,
                                           min_pids_per_host=2))

    @neutron.skip_if_missing_networking_extensions('l3-ha')
    @neutron.skip_if_missing_networking_extensions('l3_agent_scheduler')
    def test_keepalived_failover(self):
        ha_router_id = self.ha_stack.network_stack.gateway_id
        self.agents = [self.wait_for_active_ha_l3_agent()]
        keepalived_pids = self.get_cmd_pids('keepalived',
                                            ha_router_id,
                                            min_pids_per_host=2)
        ping.ping_until_received(self.ha_stack.ip_address).assert_replied()
        active_agent_host = self.agents[0]['host']

        # Need to make sure that 'keepalived-state-change' process is UP
        # before we will kill 'keepalived' process as it can break the agent
        # status otherwise. So will check that keepalived pids are equal for
        # two attemts of listing them
        ka_state_cmd = f'neutron-keepalived-state-change.*{ha_router_id}'
        ka_state_pids = {}
        for _ in tobiko.retry(timeout=120., interval=5.):
            new_ka_state_pids = self.get_cmd_pids('/usr/bin/python',
                                                  ka_state_cmd,
                                                  min_pids_per_host=1)
            if ka_state_pids == new_ka_state_pids:
                break
            else:
                ka_state_pids = new_ka_state_pids

        self.kill_pids(active_agent_host, keepalived_pids[active_agent_host])
        ping.ping_until_received(self.ha_stack.ip_address).assert_replied()

        # Need to make sure that 'keepalived' is spawned back after it has
        # been killed
        self.assertNotEqual(keepalived_pids,
                            self.get_cmd_pids('keepalived',
                                              ha_router_id,
                                              min_pids_per_host=2))

    @neutron.skip_if_missing_networking_extensions('l3_agent_scheduler')
    def test_metadata_haproxy_during_stop_L3_agent(self):
        self.agents = neutron.list_l3_agent_hosting_routers(self.router_id)
        pids = self.get_cmd_pids("haproxy", self.router_id)
        self.stop_agent()

        # Now check if haproxy processes are still run and have got same pids
        # like before dhcp agent's stop
        self.assertEqual(pids, self.get_cmd_pids("haproxy", self.router_id))

        self.start_agent()

        # And finally check if haproxy processes are still run and have got
        # same pids like at the beginning of the test
        self.assertEqual(pids, self.get_cmd_pids("haproxy", self.router_id))

    def _is_radvd_process_expected(self):
        stateless_modes = ['slaac', 'dhcpv6-stateless']
        ipv6_ra_mode = self.stack.network_stack.ipv6_subnet_details.get(
            'ipv6_ra_mode')
        ipv6_address_mode = self.stack.network_stack.ipv6_subnet_details.get(
            'ipv6_address_mode')
        if not self.stack.network_stack.ipv6_cidr:
            return False
        if (ipv6_ra_mode not in stateless_modes or
                ipv6_address_mode not in stateless_modes):
            return False
        return True

    def test_radvd_during_stop_l3_agent(self):
        os_topology = topology.get_openstack_topology()
        if os_topology.has_containers:
            self.skip("Radvd process is currently run directly in "
                      "neutron-l3-agent container so it will be always killed "
                      "when neutron-l3-agent container is killed and this "
                      "test is not needed")

        if not self._is_radvd_process_expected():
            self.skip("Radvd process is not expected to be run on router %s" %
                      self.router_id)

        self.agents = neutron.list_l3_agent_hosting_routers(self.router_id)
        pids = self.get_cmd_pids("radvd", self.router_id)
        self.stop_agent()
        # Now check if radvd processes are still run and have got same pids
        # like before dhcp agent's stop
        self.assertEqual(pids, self.get_cmd_pids("radvd", self.router_id))

        self.start_agent()

        # And finally check if dnsmasq processes are still run and have got
        # same pids like at the beginning of the test
        self.assertEqual(pids, self.get_cmd_pids("radvd", self.router_id))


class OpenVSwitchAgentTest(BaseAgentTest):

    agent_name = neutron.OPENVSWITCH_AGENT

    #: Resources stack with Nova server to send messages to
    stack = tobiko.required_setup_fixture(stacks.CirrosServerStackFixture)

    def get_agent_from_host(self, hypervisor_host):
        # pylint: disable=not-an-iterable
        short_name = tobiko.get_short_hostname(hypervisor_host)
        for agent in self.agents:
            if short_name == tobiko.get_short_hostname(agent['host']):
                return agent
        raise neutron.AgentNotFoundOnHost(agent_type=neutron.OPENVSWITCH_AGENT,
                                          host=hypervisor_host)

    def test_vm_reachability_during_stop_ovs_agent(self):
        # Check if vm is reachable before stopping service
        self.start_agent()
        ping.ping_until_received(self.stack.ip_address).assert_replied()

        # Check if vm is reachable after stopping service
        self.stop_agent(hosts=[self.stack.hypervisor_host])
        ping.ping_until_received(self.stack.ip_address).assert_replied()


class OvnControllerTest(BaseAgentTest):

    agent_name = neutron.OVN_CONTROLLER

    #: Resources stack with Nova server to send messages to
    stack = tobiko.required_setup_fixture(stacks.CirrosServerStackFixture)

    def setUp(self):
        super(OvnControllerTest, self).setUp()
        self.get_ovn_agents_from_containers()

    def kill_ovn_controller(self,
                            hosts: typing.Optional[typing.List[str]] = None,
                            timeout=60, interval=5):
        '''Stop OVN controller container by killing ovn-controller process
        running into it

        Docker/Podman service should restart it automatically

        :parm hosts: List of hostnames to stop agent on
        :type hosts: list of strings
        :param timeout: Time to wait OVN controller is recovered
        :type timeout: int
        :param interval: Time to wait between attempts
        :type interval: int
        '''
        hosts = hosts or self.hosts
        self.assertNotEqual([], hosts, "Host list is empty")

        if self.container_name == '':
            self.container_name = topology.get_agent_container_name(
                self.agent_name)

        for host in hosts:
            ssh_client = topology.get_openstack_node(hostname=host).ssh_client
            pid = None
            for directory in ('ovn', 'openvswitch'):
                try:
                    pid = sh.execute(f'{self.container_runtime_name} exec '
                                     f'-uroot {self.container_name} cat '
                                     f'/run/{directory}/ovn-controller.pid',
                                     ssh_client=ssh_client,
                                     sudo=True).stdout.splitlines()[0]
                except sh.ShellCommandFailed:
                    LOG.debug(f'/run/{directory}/ovn-controller.pid cannot '
                              f'be accessed')
                else:
                    LOG.debug(f'/run/{directory}/ovn-controller.pid returned '
                              f'pid {pid}')
                    break

            self.assertIsNotNone(pid)
            LOG.debug(f'Killing process {pid} from container '
                      f'{self.container_name} on host {host}')
            sh.execute(f'{self.container_runtime_name} exec -uroot '
                       f'{self.container_name} kill {pid}',
                       ssh_client=ssh_client,
                       sudo=True)
            LOG.debug(f'Container {self.container_name} has been killed '
                      f"on host '{host}'...")
            # Schedule auto-restart of service at the end of this test case
            self.addCleanup(self.start_agent, hosts=[host, ])

            # Verify the container is restarted automatically
            for attempt in tobiko.retry(timeout=timeout, interval=interval):
                search_running_ovn_cont = (f"{self.container_runtime_name} ps "
                                           "--format '{{.Names}}'"
                                           f" -f name={self.container_name}")
                output = sh.execute(search_running_ovn_cont,
                                    ssh_client=ssh_client,
                                    sudo=True).stdout.splitlines()

                if self.container_name in output:
                    LOG.debug(f'{self.container_name} successfully restarted')
                    break
                attempt.check_limits()

    def test_restart_ovn_controller(self):
        '''Test that OVN controller agents can be restarted successfully
        '''
        self.stop_agent()
        ping.ping_until_received(self.stack.ip_address).assert_replied()

        self.start_agent()
        ping.ping_until_received(self.stack.ip_address).assert_replied()

    def test_kill_ovn_controller(self):
        '''Test that OVN controller container is restarted automatically after
        ovn-controller process running into it was killed
        '''
        self.kill_ovn_controller()

    def test_restart_ovn_controller_containers(self):
        '''Test that OVN controller containers can be restarted successfully
        '''
        self.restart_agent_container()
        ping.ping_until_received(self.stack.ip_address).assert_replied()


class MetadataAgentTest(BaseAgentTest):

    #: Resources stack with Nova server to send messages to
    stack = tobiko.required_setup_fixture(stacks.CirrosServerStackFixture)

    agent_name = neutron.METADATA_AGENT

    def wait_for_metadata_status(self, count=None, timeout=60., interval=2.,
                                 is_reachable: typing.Optional[bool] = None):
        for attempt in tobiko.retry(timeout=timeout, interval=interval,
                                    count=count):
            if is_reachable is not None:
                try:
                    self.assert_metadata_is_reachable(is_reachable)
                except self.failureException:
                    # re-raises failureException when reaching retry limits
                    attempt.check_limits()
                else:
                    break

    def assert_metadata_is_reachable(self, is_reachable: bool,
                                     metadata_url: str = None):
        """Test if metadata agent is acting as proxy to nova metadata

        Expected response code from metadata agent is "HTTP/1.1 200 OK"
        if the agent is working. "HTTP/1.0 503 Service Unavailable" or
        exit_status=7 otherwise.
        All other HTTP statuses and exit codes are considered failures.
        """
        if is_reachable not in [True, False]:
            raise TypeError("'is_reachable' parameter is not a bool: "
                            f"{is_reachable!r}")
        # TODO: fix hard coded IP address
        metadata_url = (metadata_url or
                        'http://169.254.169.254/latest/meta-data/')

        try:
            result = sh.execute(f"curl '{metadata_url}' -I",
                                ssh_client=self.stack.ssh_client)
        except sh.ShellCommandFailed as ex:
            # Cant reach the server
            self.assertFalse(is_reachable,
                             "Metadata server not reached from Nova server:\n"
                             f"exit_status={ex.exit_status}\n"
                             f"{ex.stderr}")
            self.assertEqual(7, ex.exit_status,
                             f"Unexpected Curl exit status: {ex.exit_status}\n"
                             f"{ex.stderr}")
        else:
            # Command has succeeded, let parse the HTTP status
            curl_output = result.stdout.strip()
            LOG.debug(f"Remote HTTP server replied:\n{curl_output}")
            http_status = parse_http_status(curl_output=curl_output)
            if is_reachable:
                self.assertEqual(
                    200, http_status,
                    "Metadata server not reached from Nova server:\n"
                    f"{curl_output}")
            else:
                self.assertEqual(
                    503, http_status,
                    "Metadata server reached from Nova server:\n"
                    f"{curl_output}")

    def test_metadata_service_restart(self):
        # Ensure service is up
        self.start_agent()
        self.wait_for_metadata_status(is_reachable=True)

        # Ensure the servive gets down
        self.stop_agent()
        self.wait_for_metadata_status(is_reachable=False)

        # Ensure service gets up again
        self.start_agent()
        self.wait_for_metadata_status(is_reachable=True)

    def test_vm_reachability_when_metadata_agent_is_down(self):
        self.stop_agent()
        self.wait_for_metadata_status(is_reachable=False)
        ping.ping_until_received(self.stack.ip_address).assert_replied()
        self.start_agent()
        self.wait_for_metadata_status(is_reachable=True)

    def test_restart_metadata_containers(self):
        self.restart_agent_container()
        self.wait_for_metadata_status(is_reachable=True)


# TODO(eolivare): these tests will always be skipped on OSP13 because 'agent
# list' requests return empty list with OVN+OSP13
# Search for the corresponding container instead of the networking agent
class OvnMetadataAgentTest(MetadataAgentTest):

    agent_name = neutron.OVN_METADATA_AGENT

    def setUp(self):
        self.get_ovn_agents_from_containers()
        super(OvnMetadataAgentTest, self).setUp()


def parse_http_status(curl_output: str) -> int:
    http_head = curl_output.split('\n', 1)[0]
    return int(http_head.split(' ', 2)[1])



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import os

import netaddr
import pandas as pd
import six
import testtools

from tobiko import config
from tobiko.openstack import nova
from tobiko import tripleo
from tobiko.tripleo import pacemaker
from tobiko.tripleo import services
from tobiko.tripleo import processes
import tobiko

CONF = config.CONF


@tripleo.skip_if_missing_overcloud
class OvercloudSshConnectionTest(testtools.TestCase):

    def test_fetch_overcloud_credentials(self):
        env = tripleo.load_overcloud_rcfile()
        self.assertTrue(env['OS_AUTH_URL'])
        self.assertTrue(env.get('OS_USERNAME') or env.get('OS_USER_ID'))
        self.assertTrue(env['OS_PASSWORD'])
        self.assertTrue(env.get('OS_TENANT_NAME') or
                        env.get('OS_PROJECT_NAME') or
                        env.get('OS_TENANT_ID') or
                        env.get('OS_PROJECT_ID'))


@tripleo.skip_if_missing_overcloud
class OvercloudNovaApiTest(testtools.TestCase):

    def test_list_overcloud_nodes(self):
        nodes = tripleo.list_overcloud_nodes()
        self.assertTrue(nodes)
        for node in nodes:
            node_ip = nova.find_server_ip_address(server=node,
                                                  check_connectivity=True)
            self.assertIsInstance(node_ip, netaddr.IPAddress)

    def test_find_overcloud_nodes(self):
        node = tripleo.find_overcloud_node()
        node_ip = nova.find_server_ip_address(server=node,
                                              check_connectivity=True)
        self.assertIsInstance(node_ip, netaddr.IPAddress)

    def test_get_overcloud_node_ip_address(self):
        overcloud_node_ip = tripleo.overcloud_node_ip_address()
        self.assertIsInstance(overcloud_node_ip, netaddr.IPAddress)

    def test_overcloud_host_config(self):
        hostname = tripleo.find_overcloud_node().name
        host_config = tobiko.setup_fixture(
            tripleo.overcloud_host_config(hostname=hostname))
        self.assertEqual(hostname, host_config.host)
        self.assertIsInstance(host_config.hostname, six.string_types)
        netaddr.IPAddress(host_config.hostname)
        self.assertEqual(CONF.tobiko.tripleo.overcloud_ssh_port,
                         host_config.port)
        self.assertEqual(CONF.tobiko.tripleo.overcloud_ssh_username,
                         host_config.username)
        key_filename = tobiko.tobiko_config_path(
            CONF.tobiko.tripleo.overcloud_ssh_key_filename)
        self.assertEqual(key_filename, host_config.key_filename)
        self.assertTrue(os.path.isfile(key_filename))
        self.assertTrue(os.path.isfile(key_filename + '.pub'))

    def test_overcloud_ssh_client_connection(self):
        hostname = tripleo.find_overcloud_node().name
        ssh_client = tripleo.overcloud_ssh_client(hostname=hostname)
        ssh_client.connect()


@tripleo.skip_if_missing_overcloud
class OvercloudPacemakerTest(testtools.TestCase):
    """
    Assert that all pacemaker resources are in
    healthy state
    """

    def test_get_pacemaker_resource_table(self):
        resource_table = pacemaker.get_pcs_resources_table()
        self.assertIsInstance(resource_table, pd.DataFrame)

    def test_pacemaker_resources_health(self):
        pcs_health = pacemaker.PacemakerResourcesStatus()
        self.assertTrue(pcs_health.all_healthy)


@tripleo.skip_if_missing_overcloud
class OvercloudServicesTest(testtools.TestCase):
    """
    Assert that a subset of overcloud services are in running state
    across the overcloud nodes
    """

    def test_get_services_resource_table(self):
        oss = services.OvercloudServicesStatus()
        self.assertIsInstance(oss.oc_services_df,
                              pd.DataFrame)

    def test_overcloud_services(self):
        oss = services.OvercloudServicesStatus()
        self.assertTrue(oss.basic_overcloud_services_running)

    def test_get_overcloud_nodes_running_pcs_resource(self):
        nodes_list = pacemaker.get_overcloud_nodes_running_pcs_resource(
            resource_type='(ocf::heartbeat:rabbitmq-cluster):',
            resource_state='Started')
        self.assertIsInstance(nodes_list, list)


@tripleo.skip_if_missing_overcloud
class OvercloudProcessesTest(testtools.TestCase):
    """
    Assert that a subset of overcloud processes are in running state
    across the overcloud nodes
    """

    def test_get_processes_resource_table(self):
        ops = processes.OvercloudProcessesStatus()
        self.assertIsInstance(ops.oc_procs_df,
                              pd.DataFrame)

    def test_overcloud_processes(self):
        ops = processes.OvercloudProcessesStatus()
        self.assertTrue(ops.basic_overcloud_processes_running)



# Copyright (c) 2019 Red Hat
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import time

from oslo_log import log

import tobiko
from tobiko import config
from tobiko.openstack import keystone
from tobiko.openstack import octavia
from tobiko.openstack import stacks
from tobiko.shell import ssh
from tobiko.shell import sh
from tobiko.tests import base


LOG = log.getLogger(__name__)

CONF = config.CONF

CURL_OPTIONS = "-f --connect-timeout 2 -g"


class OctaviaOtherServerStackFixture(
        stacks.OctaviaServerStackFixture):
    pass


class OctaviaOtherMemberServerStackFixture(
        stacks.OctaviaMemberServerStackFixture):
    server_stack = tobiko.required_setup_fixture(
        OctaviaOtherServerStackFixture)


class RequestException(tobiko.TobikoException):
    message = ("Error while sending request to server "
               "(command was '{command}'): {error}")


class TimeoutException(tobiko.TobikoException):
    message = "Timeout exception: {reason}"


@keystone.skip_if_missing_service(name='octavia')
class OctaviaBasicTrafficScenarioTest(base.TobikoTest):
    """Octavia traffic scenario test.

    Create a load balancer with 2 members that run a server application,
    Create a client that is connected to the load balancer VIP port,
    Generate network traffic from the client to the load balanacer.
    """
    loadbalancer_stack = tobiko.required_setup_fixture(
        stacks.OctaviaLoadbalancerStackFixture)

    listener_stack = tobiko.required_setup_fixture(
        stacks.OctaviaListenerStackFixture)

    member1_stack = tobiko.required_setup_fixture(
        stacks.OctaviaMemberServerStackFixture)

    member2_stack = tobiko.required_setup_fixture(
        OctaviaOtherMemberServerStackFixture)

    client_stack = tobiko.required_setup_fixture(
        stacks.OctaviaClientServerStackFixture)

    members_count = 2

    def setUp(self):
        # pylint: disable=no-member
        super(OctaviaBasicTrafficScenarioTest, self).setUp()

        # Wait for members
        self._check_member(self.member1_stack)
        self._check_member(self.member2_stack)

        # Check if load balancer is functional
        self._check_loadbalancer()

    def _request(self, client_stack, server_ip_address, protocol, server_port):
        """Perform a request on a server.

        Returns the response in case of success, throws an RequestException
        otherwise.
        """
        if ':' in server_ip_address:
            # Add square brackets around IPv6 address to please curl
            server_ip_address = "[{}]".format(server_ip_address)
        cmd = "curl {} {}://{}:{}/id".format(
            CURL_OPTIONS, protocol.lower(), server_ip_address, server_port)

        ssh_client = ssh.ssh_client(
            client_stack.floating_ip_address,
            username=client_stack.image_fixture.username)

        ret = sh.ssh_execute(ssh_client, cmd)
        if ret.exit_status != 0:
            raise RequestException(command=cmd,
                                   error=ret.stderr)

        return ret.stdout

    def _wait_resource_operating_status(self, resource_type, operating_status,
                                        resource_get, *args):
        start = time.time()

        while time.time() - start < CONF.tobiko.octavia.check_timeout:
            res = resource_get(*args)
            if res['operating_status'] == operating_status:
                return

            time.sleep(CONF.tobiko.octavia.check_interval)

        raise TimeoutException(
            reason=("Cannot get operating_status '{}' from {} {} "
                    "within the timeout period.".format(
                        operating_status, resource_type, args)))

    def _wait_lb_operating_status(self, lb_id, operating_status):
        LOG.debug("Wait for loadbalancer {} to have '{}' "
                  "operating_status".format(lb_id, operating_status))
        self._wait_resource_operating_status("loadbalancer",
                                             operating_status,
                                             octavia.get_loadbalancer,
                                             lb_id)

    def _wait_for_request_data(self, client_stack, server_ip_address,
                               server_protocol, server_port):
        """Wait until a request on a server succeeds

        Throws a TimeoutException after CONF.tobiko.octavia.check_timeout
        if the server doesn't reply.
        """
        start = time.time()

        while time.time() - start < CONF.tobiko.octavia.check_timeout:
            try:
                ret = self._request(client_stack, server_ip_address,
                                    server_protocol, server_port)
            except Exception as e:
                LOG.warning("Received exception {} while performing a "
                            "request".format(e))
            else:
                return ret
            time.sleep(CONF.tobiko.octavia.check_interval)

        raise TimeoutException(
            reason=("Cannot get data from {} on port {} with "
                    "protocol {} within the timeout period.".format(
                        server_ip_address, server_port,
                        server_protocol)))

    def _check_loadbalancer(self):
        """Wait until the load balancer is functional."""

        # Check load balancer status
        loadbalancer_id = self.loadbalancer_stack.loadbalancer_id
        self._wait_lb_operating_status(loadbalancer_id, 'ONLINE')

        loadbalancer_vip = self.loadbalancer_stack.loadbalancer_vip
        loadbalancer_port = self.listener_stack.lb_port
        loadbalancer_protocol = self.listener_stack.lb_protocol

        self._wait_for_request_data(self.client_stack,
                                    loadbalancer_vip,
                                    loadbalancer_protocol,
                                    loadbalancer_port)

    def _check_member(self, member_stack):
        """Wait until a member server is functional."""

        member_ip = member_stack.server_stack.floating_ip_address
        member_port = member_stack.application_port
        member_protocol = self.listener_stack.pool_protocol

        self._wait_for_request_data(self.client_stack, member_ip,
                                    member_protocol, member_port)

    def _check_members_balanced(self):
        """Check if traffic is properly balanced between members."""
        replies = {}

        loadbalancer_vip = self.loadbalancer_stack.loadbalancer_vip
        loadbalancer_port = self.listener_stack.lb_port
        loadbalancer_protocol = self.listener_stack.lb_protocol

        for _ in range(20):
            content = self._request(self.client_stack, loadbalancer_vip,
                                    loadbalancer_protocol, loadbalancer_port)

            if content not in replies:
                replies[content] = 0
            replies[content] += 1

            # wait one second (required when using cirros' nc fake webserver)
            time.sleep(1)

        LOG.debug("Replies from load balancer: {}".format(
            replies))

        # assert that 'members_count' servers replied
        self.assertEqual(len(replies), self.members_count)

        if self.listener_stack.lb_algorithm == 'ROUND_ROBIN':
            # assert that requests have been fairly dispatched (each server
            # received the same number of requests)
            self.assertEqual(len(set(replies.values())), 1)

    def test_traffic(self):
        self._check_members_balanced()



# Copyright (c) 2019 Red Hat
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import testtools

import tobiko
from tobiko import config
from tobiko.openstack import stacks
from tobiko.openstack import neutron
from tobiko.openstack import nova
from tobiko.openstack import tests


class ServerStackResourcesTest(testtools.TestCase):

    stack = tobiko.required_setup_fixture(stacks.CirrosServerStackFixture)

    def test_server(self):
        "Test actual server details"
        server = self.stack.resources.server
        self.assertEqual('OS::Nova::Server', server.resource_type)
        # Verify actual server status (is alive, etc.)
        nova_client = nova.get_nova_client()
        server_data = nova_client.servers.get(server.physical_resource_id)
        self.assertEqual(self.stack.server_name, server_data.name)

    def test_port(self):
        "Test actual server port details"
        port = self.stack.resources.port
        self.assertEqual('OS::Neutron::Port', port.resource_type)
        # Verify actual port status (is alive, etc.)
        port_data = neutron.get_port(port.physical_resource_id)
        self.assertEqual(port.physical_resource_id, port_data['id'])


class ServerCreationTest(testtools.TestCase):

    def setUp(self):
        testtools.TestCase.setUp(self)
        if config.get_bool_env('TOBIKO_PREVENT_CREATE'):
            self.skip('TOBIKO_PREVENT_CREATE is true')

    def test_server_creation(self):
        tests.test_server_creation()

    def test_evacuable_server_creation(self):
        tests.test_evacuable_server_creation()

    def test_server_creation_and_shutoff(self):
        tests.test_server_creation_and_shutoff()

    def test_servers_creation(self):
        tests.test_servers_creation()



# Copyright (c) 2019 Red Hat
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import


import testtools

import tobiko
from tobiko.openstack import neutron
from tobiko.openstack import nova
from tobiko.openstack import stacks
from tobiko.shell import ping
from tobiko.shell import sh


class NetworkTest(testtools.TestCase):

    #: Resources stack with Nova server to send messages to
    stack = tobiko.required_setup_fixture(stacks.CirrosPeerServerStackFixture)

    def test_stack_create_complete(self):
        self.stack.key_pair_stack.wait_for_create_complete()
        self.stack.network_stack.wait_for_create_complete()
        self.stack.peer_stack.wait_for_create_complete()
        self.stack.wait_for_create_complete()

    def test_ssh(self):
        """Test SSH connectivity to floating IP address"""
        hostname = sh.get_hostname(ssh_client=self.stack.ssh_client)
        self.assertEqual(self.stack.server_name.lower(), hostname)

    def test_ping(self):
        """Test ICMP connectivity to floating IP address"""
        ping.ping_until_received(
            self.stack.ip_address,
            ssh_client=self.stack.peer_stack.ssh_client).assert_replied()

    # --- test l3_ha extension ------------------------------------------------

    @neutron.skip_if_missing_networking_extensions('l3-ha')
    def test_l3_ha(self):
        """Test 'mtu' network attribute"""
        gateway = self.stack.network_stack.gateway_details
        self.assertEqual(self.stack.network_stack.ha,
                         gateway['ha'])


@tobiko.skip("The test is not able to allocate required resources")
class SameHostNetworkTest(NetworkTest):

    #: Resources stack with Nova server to send messages to
    stack = tobiko.required_setup_fixture(
        stacks.CirrosSameHostServerStackFixture)

    def test_same_host(self):
        sender = self.stack.peer_stack.server_details
        receiver = self.stack.server_details
        self.assertEqual({'same_host': [sender.id]},
                         self.stack.scheduler_hints)
        self.assertEqual(getattr(sender, 'OS-EXT-SRV-ATTR:host'),
                         getattr(receiver, 'OS-EXT-SRV-ATTR:host'))


@tobiko.skip("The test is not able to allocate required resources")
@nova.skip_if_missing_hypervisors(count=2, state='up', status='enabled')
class DifferentHostNetworkTest(NetworkTest):

    #: Resources stack with Nova server to send messages to
    stack = tobiko.required_setup_fixture(
        stacks.CirrosDifferentHostServerStackFixture)

    def test_different_host(self):
        sender = self.stack.peer_stack.server_details
        receiver = self.stack.server_details
        self.assertEqual({'different_host': [sender.id]},
                         self.stack.scheduler_hints)
        self.assertNotEqual(getattr(sender, 'OS-EXT-SRV-ATTR:host'),
                            getattr(receiver, 'OS-EXT-SRV-ATTR:host'))


# --- l3-ha extension VM to VM scenario ---------------------------------------

@neutron.skip_if_missing_networking_extensions('l3-ha')
@neutron.skip_if_missing_networking_agents(binary='neutron-l3-agent',
                                           count=2)
class L3haNetworkTest(NetworkTest):
    #: Resources stack with floating IP and Nova server
    stack = tobiko.required_setup_fixture(stacks.L3haPeerServerStackFixture)


@neutron.skip_if_missing_networking_extensions('l3-ha')
@neutron.skip_if_missing_networking_agents(binary='neutron-l3-agent',
                                           count=2)
class L3haSameHostNetworkTest(SameHostNetworkTest):
    #: Resources stack with Nova server to send messages to
    stack = tobiko.required_setup_fixture(
        stacks.L3haSameHostServerStackFixture)


@neutron.skip_if_missing_networking_extensions('l3-ha')
@neutron.skip_if_missing_networking_agents(binary='neutron-l3-agent',
                                           count=2)
class L3haDifferentHostNetworkTest(DifferentHostNetworkTest):
    #: Resources stack with Nova server to send messages to
    stack = tobiko.required_setup_fixture(
        stacks.L3haDifferentHostServerStackFixture)



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from tobiko.docker import _client
from tobiko.docker import _shell
from tobiko.docker import _exception


DockerClientFixture = _client.DockerClientFixture
get_docker_client = _client.get_docker_client
list_docker_containers = _client.list_docker_containers

discover_docker_urls = _shell.discover_docker_urls
is_docker_running = _shell.is_docker_running

DockerError = _exception.DockerError
DockerUrlNotFoundError = _exception.DockerUrlNotFoundError



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import bz2
import gzip
import io
import zipfile

from oslo_log import log

from tobiko.openstack.glance import _lzma


LOG = log.getLogger(__name__)

COMPRESSED_FILE_TYPES = {}


def comressed_file_type(cls):
    COMPRESSED_FILE_TYPES[cls.compression_type] = cls
    return cls


@comressed_file_type
class BZ2FileType(object):
    file_magic = b'\x42\x5a\x68'
    compression_type = 'bz2'
    open_file = bz2.BZ2File


@comressed_file_type
class GzipFileType(gzip.GzipFile):
    file_magic = b'\x1f\x8b\x08'
    compression_type = 'gz'
    open_file = gzip.GzipFile


@comressed_file_type
class XzFileType(object):
    file_magic = b'\xfd7zXZ\x00'
    compression_type = 'xz'
    open_file = staticmethod(_lzma.open_file)


@comressed_file_type
class ZipFileType(object):
    file_magic = b'\x50\x4b\x03\x04'
    compression_type = 'zip'
    open_file = zipfile.ZipFile


def open_image_file(filename, mode, compression_type=None):
    if compression_type is None:
        max_magic_len = max(len(cls.file_magic)
                            for cls in COMPRESSED_FILE_TYPES.values())
        with io.open(filename, 'rb') as f:
            magic = f.read(max_magic_len)
        for cls in COMPRESSED_FILE_TYPES.values():
            if magic.startswith(cls.file_magic):
                compression_type = cls.compression_type
                LOG.debug("Compression type %r of file %r got from file magic",
                          compression_type, filename)
                break

    if compression_type:
        LOG.debug("Open compressed file %r (mode=%r, compression_type=%r)",
                  filename, mode, compression_type)
        open_func = COMPRESSED_FILE_TYPES[compression_type].open_file
    else:
        LOG.debug("Open flat file %r (mode=%r)", filename, mode)
        open_func = io.open

    return open_func(filename, mode)



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import collections
import typing

import netaddr
from neutronclient.v2_0 import client as neutronclient

import tobiko
from tobiko.openstack import _client


ServiceUnavailable = neutronclient.exceptions.ServiceUnavailable


class NeutronClientFixture(_client.OpenstackClientFixture):

    def init_client(self, session):
        return neutronclient.Client(session=session)


class NeutronClientManager(_client.OpenstackClientManager):

    def create_client(self, session):
        return NeutronClientFixture(session=session)


CLIENTS = NeutronClientManager()


NeutronClientType = typing.Union[None,
                                 neutronclient.Client,
                                 NeutronClientFixture]


def neutron_client(obj: NeutronClientType) -> neutronclient.Client:
    if not obj:
        return get_neutron_client()

    if isinstance(obj, neutronclient.Client):
        return obj

    fixture = tobiko.setup_fixture(obj)
    if isinstance(fixture, NeutronClientFixture):
        return fixture.client

    message = "Object {!r} is not a NeutronClientFixture".format(obj)
    raise TypeError(message)


class HasNeutronClientFixture(tobiko.SharedFixture):

    client: typing.Optional[neutronclient.Client] = None

    def __init__(self, obj: NeutronClientType = None):
        super(HasNeutronClientFixture, self).__init__()
        self._obj = obj

    def setup_fixture(self):
        self.client = neutron_client(self._obj)


def get_neutron_client(session=None, shared=True, init_client=None,
                       manager=None) -> neutronclient.Client:
    manager = manager or CLIENTS
    client = manager.get_client(session=session, shared=shared,
                                init_client=init_client)
    tobiko.setup_fixture(client)
    return client.client


_RAISE_ERROR = object()


def find_port(client=None, unique=False, default=_RAISE_ERROR, **attributes):
    """Look for a port matching some property values"""
    ports = list_ports(client=client, **attributes)
    if default is _RAISE_ERROR or ports:
        if unique:
            return ports.unique
        else:
            return ports.first
    else:
        return default


def find_subnet(client=None, unique=False, default=_RAISE_ERROR, **attributes):
    """Look for a subnet matching some property values"""
    subnets = list_subnets(client=client, **attributes)
    if default is _RAISE_ERROR or subnets:
        if unique:
            return subnets.unique
        else:
            return subnets.first
    else:
        return default


def list_ports(client=None, **params):
    ports = neutron_client(client).list_ports(**params)['ports']
    return tobiko.select(ports)


NeutronSubnetType = typing.Dict[str, typing.Any]


def list_subnets(client=None, ip_version: typing.Optional[int] = None,
                 **params) -> tobiko.Selection[NeutronSubnetType]:
    if ip_version is not None:
        params['ip_version'] = ip_version
    subnets = neutron_client(client).list_subnets(**params)
    if isinstance(subnets, collections.Mapping):
        subnets = subnets['subnets']
    return tobiko.select(subnets)


def list_agents(client=None, **params):
    agents = neutron_client(client).list_agents(**params)
    if isinstance(agents, collections.Mapping):
        agents = agents['agents']
    return tobiko.select(agents)


def list_subnet_cidrs(client=None, **params):
    return tobiko.select(netaddr.IPNetwork(subnet['cidr'])
                         for subnet in list_subnets(client=client, **params))


def get_floating_ip(floating_ip, client=None, **params):
    floating_ip = neutron_client(client).show_floatingip(floating_ip, **params)
    return floating_ip['floatingip']


def get_port(port, client=None, **params):
    try:
        return neutron_client(client).show_port(port, **params)['port']
    except neutronclient.exceptions.NotFound as ex:
        raise NoSuchPort(id=port) from ex


def get_router(router, client=None, **params):
    try:
        return neutron_client(client).show_router(router, **params)['router']
    except neutronclient.exceptions.NotFound as ex:
        raise NoSuchRouter(id=router) from ex


def get_subnet(subnet, client=None, **params):
    try:
        return neutron_client(client).show_subnet(subnet, **params)['subnet']
    except neutronclient.exceptions.NotFound as ex:
        raise NoSuchSubnet(id=subnet) from ex


def list_l3_agent_hosting_routers(router, client=None, **params):
    agents = neutron_client(client).list_l3_agent_hosting_routers(
        router, **params)
    if isinstance(agents, collections.Mapping):
        agents = agents['agents']
    return tobiko.select(agents)


def find_l3_agent_hosting_router(router, client=None, unique=False,
                                 default=_RAISE_ERROR, **params):
    agents = list_l3_agent_hosting_routers(router=router, client=client,
                                           **params)
    if default is _RAISE_ERROR or agents:
        if unique:
            return agents.unique
        else:
            return agents.first
    else:
        return default


def list_dhcp_agent_hosting_network(network, client=None, **params):
    agents = neutron_client(client).list_dhcp_agent_hosting_networks(
        network, **params)
    if isinstance(agents, collections.Mapping):
        agents = agents['agents']
    return tobiko.select(agents)


class NoSuchPort(tobiko.ObjectNotFound):
    message = "No such port found for {id!r}"


class NoSuchRouter(tobiko.ObjectNotFound):
    message = "No such router found for {id!r}"


class NoSuchSubnet(tobiko.ObjectNotFound):
    message = "No such subnet found for {id!r}"



# Copyright (c) 2019 Red Hat
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import testtools

import tobiko
from tobiko import config
from tobiko.shell import ping
from tobiko.shell import sh
from tobiko.openstack import neutron
from tobiko.openstack import stacks


CONF = config.CONF


class FloatingIPTest(testtools.TestCase):
    """Tests connectivity via floating IPs"""

    #: Resources stack with floating IP and Nova server
    stack = tobiko.required_setup_fixture(stacks.CirrosServerStackFixture)

    def test_stack_create_complete(self):
        self.stack.key_pair_stack.wait_for_create_complete()
        self.stack.network_stack.wait_for_create_complete()
        self.stack.wait_for_create_complete()

    def test_ssh(self):
        """Test SSH connectivity to floating IP address"""
        hostname = sh.get_hostname(ssh_client=self.stack.ssh_client)
        self.assertEqual(self.stack.server_name.lower(), hostname)

    def test_ping(self):
        """Test ICMP connectivity to floating IP address"""
        ping.ping_until_received(
            self.stack.floating_ip_address).assert_replied()

    # --- test port-security extension ---------------------------------------

    @neutron.skip_if_missing_networking_extensions('port-security')
    def test_port_security_enabled_port_attribute(self):
        """Test port security enabled port attribute"""
        self.assertEqual(self.expected_port_security_enabled,
                         self.observed_port_security_enabled)

    @property
    def expected_port_security_enabled(self):
        """Expected port security enabled value"""
        return self.stack.port_security_enabled

    @property
    def observed_port_security_enabled(self):
        """Actual MTU value for internal network"""
        return self.stack.outputs.port_security_enabled

    # --- test security_group extension --------------------------------------

    @neutron.skip_if_missing_networking_extensions('security-group')
    def test_security_groups_port_attribute(self):
        """Test security groups port attribute"""
        self.assertEqual(self.expected_security_groups,
                         self.observed_security_groups)

    @property
    def expected_security_groups(self):
        """Expected port security groups"""
        return set(self.stack.security_groups)

    @property
    def observed_security_groups(self):
        """Actual port security group"""
        return set(self.stack.outputs.security_groups)

    # --- test net-mtu and net-mtu-writable extensions ------------------------

    @ping.skip_if_missing_fragment_ping_option
    @neutron.skip_if_missing_networking_extensions('net-mtu')
    def test_ping_with_net_mtu(self):
        """Test connectivity to floating IP address with MTU sized packets"""
        # Wait until it can reach remote port with maximum-sized packets
        ping.ping(self.stack.floating_ip_address,
                  until=ping.RECEIVED,
                  packet_size=self.observed_net_mtu,
                  fragmentation=False).assert_replied()

        # Verify it can't reach remote port with over-sized packets
        ping.ping(self.stack.floating_ip_address,
                  packet_size=self.observed_net_mtu + 1,
                  fragmentation=False,
                  count=5,
                  check=False).assert_not_replied()

    @property
    def observed_net_mtu(self):
        """Actual MTU value for internal network"""
        return self.stack.network_stack.outputs.mtu

    # --- test l3_ha extension ------------------------------------------------

    @neutron.skip_if_missing_networking_extensions('l3-ha')
    def test_l3_ha(self):
        """Test 'mtu' network attribute"""
        gateway = self.stack.network_stack.gateway_details
        self.assertEqual(self.stack.network_stack.ha,
                         gateway['ha'])


# --- Test with port security enabled -----------------------------------------

@neutron.skip_if_missing_networking_extensions('port-security',
                                               'security-group')
class FloatingIPWithPortSecurityFixture(stacks.CirrosServerStackFixture):
    """Heat stack for testing a floating IP instance with port security"""

    #: Resources stack with security group to allow ping Nova servers
    security_groups_stack = tobiko.required_setup_fixture(
        stacks.SecurityGroupsFixture)

    #: Enable port security on internal network
    port_security_enabled = True

    @property
    def security_groups(self):
        """List with ICMP security group"""
        return [self.security_groups_stack.ssh_security_group_id]


@neutron.skip_if_missing_networking_extensions('port-security',
                                               'security-group')
class FloatingIPWithPortSecurityTest(FloatingIPTest):
    """Tests connectivity via floating IPs with port security"""

    #: Resources stack with floating IP and Nova server with port security
    stack = tobiko.required_setup_fixture(FloatingIPWithPortSecurityFixture)

    def test_ping(self):
        """Test connectivity to floating IP address"""
        # Wait for server instance to get ready by logging in
        self.stack.ssh_client.connect()

        # Check can't reach secured port via floating IP
        ping.ping(self.stack.floating_ip_address,
                  count=5,
                  check=False).assert_not_replied()

    @ping.skip_if_missing_fragment_ping_option
    @neutron.skip_if_missing_networking_extensions('net-mtu')
    def test_ping_with_net_mtu(self):
        """Test connectivity to floating IP address"""
        # Wait for server instance to get ready by logging in
        tobiko.setup_fixture(self.stack.ssh_client)
        self.stack.ssh_client.connect()

        # Verify it can't reach secured port with maximum-sized packets
        ping.ping(self.stack.floating_ip_address,
                  packet_size=self.observed_net_mtu,
                  fragmentation=False,
                  count=5,
                  check=False).assert_not_replied()

        # Verify it can't reach secured port with over-sized packets
        ping.ping(self.stack.floating_ip_address,
                  packet_size=self.observed_net_mtu + 1,
                  fragmentation=False,
                  count=5,
                  check=False).assert_not_replied()


# --- Test with ICMP security group -------------------------------------------

class FloatingIPWithICMPSecurityGroupFixture(
        FloatingIPWithPortSecurityFixture):
    """Heat stack for testing a floating IP instance with security groups"""

    @property
    def security_groups(self):
        """List with ICMP security group"""
        return [self.security_groups_stack.ssh_security_group_id,
                self.security_groups_stack.icmp_security_group_id]


@neutron.skip_if_missing_networking_extensions('port-security',
                                               'security-group')
class FloatingIPWithICMPSecurityGroupTest(FloatingIPTest):
    """Tests connectivity via floating IP with security ICMP security group"""
    #: Resources stack with floating IP and Nova server to ping
    stack = tobiko.required_setup_fixture(
        FloatingIPWithICMPSecurityGroupFixture)


# --- Test net-mtu-write extension --------------------------------------------

@neutron.skip_if_missing_networking_extensions('net-mtu-writable')
class FloatingIPWithNetMtuWritableFixture(stacks.CirrosServerStackFixture):
    """Heat stack for testing floating IP with a custom MTU network value"""

    #: Heat stack for creating internal network with custom MTU value
    network_stack = tobiko.required_setup_fixture(
        stacks.NetworkWithNetMtuWriteStackFixture)


@neutron.skip_if_missing_networking_extensions('net-mtu-writable')
class FloatingIpWithMtuWritableTest(FloatingIPTest):
    """Tests connectivity via floating IP with a custom MTU value"""

    #: Resources stack with floating IP and Nova server
    stack = tobiko.required_setup_fixture(FloatingIPWithNetMtuWritableFixture)

    def test_net_mtu_write(self):
        """Test 'mtu' network attribute"""
        self.assertEqual(self.expected_net_mtu, self.observed_net_mtu)

    @property
    def expected_net_mtu(self):
        """Expected MTU value for internal network"""
        return self.stack.network_stack.custom_mtu_size


# --- Test la-h3 extension ----------------------------------------------------

@neutron.skip_if_missing_networking_extensions('l3-ha')
@neutron.skip_if_missing_networking_agents(binary='neutron-l3-agent',
                                           count=2)
class FloatingIpWithL3HATest(FloatingIPTest):
    #: Resources stack with floating IP and Nova server
    stack = tobiko.required_setup_fixture(stacks.L3haServerStackFixture)



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import collections
import typing  # noqa
import weakref

import netaddr
from oslo_log import log
import six
from six.moves.urllib import parse

import tobiko
from tobiko import docker
from tobiko import podman
from tobiko.shell import files
from tobiko.shell import ip
from tobiko.shell import sh
from tobiko.shell import ssh
from tobiko.openstack import neutron
from tobiko.openstack import nova
from tobiko.openstack import keystone
from tobiko.openstack.topology import _address
from tobiko.openstack.topology import _config
from tobiko.openstack.topology import _connection
from tobiko.openstack.topology import _exception


LOG = log.getLogger(__name__)


def list_openstack_nodes(topology=None, group=None, hostnames=None, **kwargs):
    topology = topology or get_openstack_topology()
    if group is None:
        nodes = topology.nodes
    elif isinstance(group, str):
        nodes = topology.get_group(group=group)
    else:
        nodes = topology.get_groups(groups=group)

    if hostnames:
        names = {node_name_from_hostname(hostname)
                 for hostname in hostnames}
        nodes = [node
                 for node in nodes
                 if node.name in names]
    if kwargs:
        nodes = nodes.with_attributes(**kwargs)
    return nodes


def find_openstack_node(topology=None, unique=False, **kwargs):
    nodes = list_openstack_nodes(topology=topology, **kwargs)
    if unique:
        return nodes.unique
    else:
        return nodes.first


def get_openstack_node(hostname, address=None, topology=None):
    topology = topology or get_openstack_topology()
    return topology.get_node(hostname=hostname, address=address)


def list_openstack_node_groups(topology=None):
    topology = topology or get_openstack_topology()
    return topology.groups


def get_default_openstack_topology_class() -> typing.Type:
    return DEFAULT_TOPOLOGY_CLASS


def set_default_openstack_topology_class(topology_class: typing.Type):
    # pylint: disable=global-statement
    if not issubclass(topology_class, OpenStackTopology):
        raise TypeError(f"'{topology_class}' is not subclass of "
                        f"'{OpenStackTopology}'")
    global DEFAULT_TOPOLOGY_CLASS
    DEFAULT_TOPOLOGY_CLASS = topology_class


def get_agent_service_name(agent_name: str) -> str:
    topology_class = get_default_openstack_topology_class()
    return topology_class.get_agent_service_name(agent_name)


def get_agent_container_name(agent_name: str) -> str:
    topology_class = get_default_openstack_topology_class()
    return topology_class.get_agent_container_name(agent_name)


def check_systemd_monitors_agent(hostname: str, agent_name: str) -> bool:
    ssh_client = get_openstack_node(hostname=hostname).ssh_client
    services = sh.execute('systemctl list-units --type=service --all '
                          '--no-pager --no-legend',
                          ssh_client=ssh_client,
                          sudo=True).stdout.split('\n')
    for service in services:
        if get_agent_service_name(agent_name) in service:
            return True
    return False


class UnknowOpenStackServiceNameError(tobiko.TobikoException):
    message = ("Unknown service name for agent name '{agent_name}' and "
               "topology class '{topology_class}'")


class UnknowOpenStackContainerNameError(tobiko.TobikoException):
    message = ("Unknown container name for agent name '{agent_name}' and "
               "topology class '{topology_class}'")


class OpenStackTopologyNode(object):

    _docker_client = None
    _podman_client = None

    def __init__(self, topology, name: str, ssh_client: ssh.SSHClientFixture,
                 addresses: typing.List[netaddr.IPAddress], hostname: str):
        self._topology = weakref.ref(topology)
        self.name = name
        self.ssh_client = ssh_client
        self.groups: typing.Set[str] = set()
        self.addresses: typing.List[netaddr.IPAddress] = list(addresses)
        self.hostname: str = hostname

    @property
    def topology(self):
        return self._topology()

    def add_group(self, group: str):
        self.groups.add(group)

    @property
    def public_ip(self):
        return self.addresses[0]

    @property
    def ssh_parameters(self):
        return self.ssh_client.setup_connect_parameters()

    @property
    def docker_client(self):
        docker_client = self._docker_client
        if not docker_client:
            self._docker_client = docker_client = docker.get_docker_client(
                ssh_client=self.ssh_client)
        return docker_client

    @property
    def podman_client(self):
        podman_client = self._podman_client
        if not podman_client:
            self._podman_client = podman_client = podman.get_podman_client(
                ssh_client=self.ssh_client)
        return podman_client

    def __repr__(self):
        return "{cls!s}<name={name!r}>".format(cls=type(self).__name__,
                                               name=self.name)


class OpenStackTopology(tobiko.SharedFixture):

    config = tobiko.required_setup_fixture(_config.OpenStackTopologyConfig)

    agent_to_service_name_mappings = {
        neutron.DHCP_AGENT: 'devstack@q-dhcp',
        neutron.L3_AGENT: 'devstack@q-l3',
        neutron.OPENVSWITCH_AGENT: 'devstack@q-agt',
        neutron.METADATA_AGENT: 'devstack@q-meta',
        neutron.OVN_METADATA_AGENT: 'devstack@q-ovn-metadata-agent',
        neutron.OVN_CONTROLLER: 'devstack@ovn-controller'
    }
    agent_to_container_name_mappings: typing.Dict[str, str] = {}

    has_containers = False

    _connections = tobiko.required_setup_fixture(
        _connection.SSHConnectionManager)

    # In Devstack based env logs can be accessed by journalctl
    file_digger_class: typing.Type
    file_digger_class = files.JournalLogDigger

    # This is dict which handles mapping of the log file and systemd_unit (if
    # needed) for the OpenStack services.
    # In case of Devstack topology file name in fact name of the systemd unit
    # as logs are stored in journalctl
    log_names_mappings = {
        neutron.SERVER: 'devstack@q-svc',
    }

    def __init__(self):
        super(OpenStackTopology, self).__init__()
        self._names: typing.Dict[str, OpenStackTopologyNode] = (
            collections.OrderedDict())
        self._groups: typing.Dict[str, tobiko.Selection] = (
            collections.OrderedDict())
        self._addresses: typing.Dict[netaddr.IPAddress,
                                     OpenStackTopologyNode] = (
            collections.OrderedDict())

    def setup_fixture(self):
        self.discover_nodes()

    def cleanup_fixture(self):
        tobiko.cleanup_fixture(self._connections)
        self._names.clear()
        self._groups.clear()
        self._addresses.clear()

    @classmethod
    def get_agent_service_name(cls, agent_name: str) -> str:
        try:
            return cls.agent_to_service_name_mappings[agent_name]
        except KeyError:
            pass
        raise UnknowOpenStackServiceNameError(agent_name=agent_name,
                                              topology_class=cls)

    @classmethod
    def get_agent_container_name(cls, agent_name: str) -> str:
        try:
            return cls.agent_to_container_name_mappings[agent_name]
        except KeyError:
            pass
        raise UnknowOpenStackContainerNameError(agent_name=agent_name,
                                                topology_class=cls)

    def discover_nodes(self):
        self.discover_configured_nodes()
        self.discover_controller_nodes()
        self.discover_compute_nodes()

    def discover_configured_nodes(self):
        for address in self.config.conf.nodes or []:
            self.add_node(address=address)

    def discover_controller_nodes(self):
        endpoints = keystone.list_endpoints(interface='public')
        addresses = set(parse.urlparse(endpoint.url).hostname
                        for endpoint in endpoints)
        for address in addresses:
            try:
                self.add_node(address=address, group='controller')
            except _connection.UreachableSSHServer as ex:
                LOG.debug(f"Unable to SSH to end point address '{address}'. "
                          f"{ex}")

    def discover_compute_nodes(self):
        for hypervisor in nova.list_hypervisors():
            self.add_node(hostname=hypervisor.hypervisor_hostname,
                          address=hypervisor.host_ip,
                          group='compute')

    def add_node(self,
                 hostname: typing.Optional[str] = None,
                 address: typing.Optional[str] = None,
                 group: typing.Optional[str] = None,
                 ssh_client: typing.Optional[ssh.SSHClientFixture] = None) \
            -> OpenStackTopologyNode:
        if ssh_client is not None:
            # detect all global addresses from remote server
            try:
                hostname = sh.get_hostname(ssh_client=ssh_client)
            except Exception:
                LOG.exception("Unable to get node hostname from "
                              f"{ssh_client}")
                ssh_client = None
        name = hostname and node_name_from_hostname(hostname) or None

        addresses: typing.List[netaddr.IPAddress] = []
        if address:
            # add manually configure addresses first
            addresses.extend(self._list_addresses(address))
        if hostname:
            # detect more addresses from the hostname
            addresses.extend(self._list_addresses(hostname))
        addresses = tobiko.select(remove_duplications(addresses))

        try:
            node = self.get_node(name=name, address=addresses)
        except _exception.NoSuchOpenStackTopologyNode:
            node = None

        node = node or self._add_node(addresses=addresses,
                                      hostname=hostname,
                                      ssh_client=ssh_client)

        if group:
            # Add group anyway even if the node hasn't been added
            group_nodes = self.add_group(group=group)
            if node and node not in group_nodes:
                group_nodes.append(node)
                node.add_group(group=group)

        return node

    def _add_node(self,
                  addresses: typing.List[netaddr.IPAddress],
                  hostname: str = None,
                  ssh_client: typing.Optional[ssh.SSHClientFixture] = None):
        if ssh_client is None:
            ssh_client = self._ssh_connect(addresses=addresses)
        addresses.extend(self._list_addresses_from_host(ssh_client=ssh_client))
        addresses = tobiko.select(remove_duplications(addresses))
        hostname = hostname or sh.get_hostname(ssh_client=ssh_client)
        name = node_name_from_hostname(hostname)
        try:
            node = self._names[name]
        except KeyError:
            LOG.debug("Add topology node:\n"
                      f" - name: {name}\n"
                      f" - hostname: {hostname}\n"
                      f" - login: {ssh_client.login}\n"
                      f" - addresses: {addresses}\n")
            self._names[name] = node = self.create_node(name=name,
                                                        hostname=hostname,
                                                        ssh_client=ssh_client,
                                                        addresses=addresses)

        for address in addresses:
            address_node = self._addresses.setdefault(address, node)
            if address_node is not node:
                LOG.error(f"Address '{address}' of node '{name}' is already "
                          f"used by node '{address_node.name}'")
        return node

    def get_node(self, name=None, hostname=None, address=None):
        name = name or (hostname and node_name_from_hostname(hostname))
        details = {}
        if name:
            tobiko.check_valid_type(name, six.string_types)
            details['name'] = name
            try:
                return self._names[name]
            except KeyError:
                pass
        if address:
            details['address'] = address
            for address in self._list_addresses(address):
                try:
                    return self._addresses[address]
                except KeyError:
                    pass
        raise _exception.NoSuchOpenStackTopologyNode(details=details)

    def create_node(self, name, ssh_client, **kwargs):
        return OpenStackTopologyNode(topology=self, name=name,
                                     ssh_client=ssh_client, **kwargs)

    @property
    def nodes(self):
        return tobiko.select(self.get_node(name)
                             for name in self._names)

    def add_group(self, group: str) -> tobiko.Selection:
        try:
            return self._groups[group]
        except KeyError:
            self._groups[group] = nodes = self.create_group()
            return nodes

    @staticmethod
    def create_group() -> tobiko.Selection[OpenStackTopologyNode]:
        return tobiko.Selection()

    def get_group(self, group) -> tobiko.Selection[OpenStackTopologyNode]:
        try:
            return self._groups[group]
        except KeyError as ex:
            raise _exception.NoSuchOpenStackTopologyNodeGroup(
                group=group) from ex

    def get_groups(self, groups) -> tobiko.Selection[OpenStackTopologyNode]:
        nodes: tobiko.Selection[OpenStackTopologyNode] = tobiko.Selection()
        for group in groups:
            nodes.extend(self.get_group(group))
        return nodes

    @property
    def groups(self) -> typing.List[str]:
        return list(self._groups)

    def _ssh_connect(self, addresses: typing.List[netaddr.IPAddress],
                     **connect_params) -> ssh.SSHClientFixture:

        try:
            return _connection.ssh_connect(addresses, **connect_params)
        except _connection.UreachableSSHServer:
            for proxy_node in self.nodes:
                proxy_client = proxy_node.ssh_client
                if proxy_client:
                    LOG.debug("Try connecting through a proxy node "
                              f"'{proxy_node.name}'")
                    try:
                        return self._ssh_connect_with_proxy_client(
                            addresses, proxy_client, **connect_params)
                    except _connection.UreachableSSHServer:
                        pass
            raise

    def _ssh_connect_with_proxy_client(self, addresses, proxy_client,
                                       **connect_params) -> \
            ssh.SSHClientFixture:
        ssh_client = _connection.ssh_connect(addresses,
                                             proxy_client=proxy_client,
                                             **connect_params)
        addresses = self._list_addresses_from_host(ssh_client=ssh_client)
        try:
            LOG.debug("Try connecting through an address that doesn't require "
                      "an SSH proxy host")
            return _connection.ssh_connect(addresses, **connect_params)
        except _connection.UreachableSSHServer:
            return ssh_client

    @property
    def ip_version(self) -> typing.Optional[int]:
        ip_version = self.config.conf.ip_version
        return ip_version and int(ip_version) or None

    def _list_addresses_from_host(self, ssh_client: ssh.SSHClientFixture):
        return ip.list_ip_addresses(ssh_client=ssh_client,
                                    ip_version=self.ip_version,
                                    scope='global')

    def _list_addresses(self, obj) -> typing.List[netaddr.IPAddress]:
        return _address.list_addresses(obj,
                                       ip_version=self.ip_version,
                                       ssh_config=True)


def get_openstack_topology(topology_class: typing.Type = None) -> \
        OpenStackTopology:
    if topology_class:
        if not issubclass(topology_class, OpenStackTopology):
            raise TypeError(f"'{topology_class}' is not subclass of "
                            f"'{OpenStackTopology}'")
    else:
        topology_class = get_default_openstack_topology_class()
    return tobiko.setup_fixture(topology_class)


DEFAULT_TOPOLOGY_CLASS = OpenStackTopology


def node_name_from_hostname(hostname):
    return hostname.split('.', 1)[0].lower()


def remove_duplications(items: typing.List) -> typing.List:
    # use all items as dictionary keys to remove duplications
    mapping = collections.OrderedDict((k, None) for k in items)
    return list(mapping.keys())



# Copyright 2020 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import


NOT_CACHED = object()


class CachedProperty(object):
    """ Property that calls getter only the first time it is required

    It invokes property setter with the value returned by getter the first time
    it is required so that it is not requested again until del operator is not
    called again on top of target object.

    Implements default setter and deleter behaving as a regular attribute would
    by setting or removing named attribute of target object __dict__

    The name used for storing cached property value is got from getter function
    name if not passed as constructor attribute.

    Examples of use:

      class MyClass(object):

          @cached
          def my_property(self):
              return object()


      obj = MyClass()
      # my_property method not yet called
      assert 'my_property' not in obj.__dict__

      # my_property method is called
      first_value = obj.my_property
      assert obj.__dict__['my_property'] is first_value

      # my_property method is not called again
      assert obj.my_property is first_value

      # first value is removed from dictionary
      del obj.my_property
      assert 'my_property' not in obj.__dict__

      # my_property method is called
      second_value = obj.my_property
      assert obj.__dict__['my_property'] is second_value

      # value returned by second call of method can be different
      second_value is not first_value

    For more details about how Python properties works please refers to
    language documentation [1]

    [1] https://docs.python.org/3/howto/descriptor.html
    """

    fget = None
    fset = None
    fdel = None
    __doc__ = None
    cached_id = None

    def __init__(self, fget=None, fset=None, fdel=None, doc=None,
                 cached_id=None):
        if fget:
            self.getter(fget)
        if fset:
            self.setter(fset)
        if fdel:
            self.deleter(fdel)
        if doc:
            self.__doc__ = doc
        if cached_id:
            self.cached_id = cached_id
        elif self.cached_id is None:
            self.cached_id = '_cached_' + str(id(self))

    def getter(self, fget):
        assert callable(fget)
        self.fget = fget
        if self.__doc__ is None:
            self.__doc__ = fget.__doc__
        return fget

    def setter(self, fset):
        self.fset = fset
        return fset

    def deleter(self, fdel):
        self.fdel = fdel
        return fdel

    def __get__(self, obj, _objtype=None):
        if obj is None:
            return self

        value = self._get_cached(obj)
        if value is NOT_CACHED:
            if self.fget is None:
                raise AttributeError("Cached property has no getter method")
            value = self.fget(obj)
            self.__set__(obj, value)

        return value

    def __set__(self, obj, value):
        if self.fset:
            self.fset(obj, value)
        self._set_cached(obj, value)

    def __delete__(self, obj):
        if self.fdel:
            self.fdel(obj)
        self._delete_cached(obj)

    def _get_cached(self, obj):
        return getattr(obj, self.cached_id, NOT_CACHED)

    def _set_cached(self, obj, value):
        setattr(obj, self.cached_id, value)

    def _delete_cached(self, obj):
        return obj.__dict__.pop(self.cached_id, NOT_CACHED)


def cached(*args, **kwargs):
    return CachedProperty(*args, **kwargs)
